# #!/usr/bin/env python
# # =============================================================================#
# #                                                                             #
# # NAME:     rmtools_bwdepol.py                                                #
# #                                                                             #
# # PURPOSE: Algorithm for finding polarized sources while accounting for       #
# #          bandwidth depolarization.                                          #
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2020 Canadian Initiative for Radio Astronomy Data Analysis     #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#

# import argparse
# import math as m
# import os
# import sys
# import time
# import traceback

# import matplotlib.pyplot as plt
# import numpy as np
# import scipy as sp
# from matplotlib.ticker import MaxNLocator
# from utils.util_plotTk import (
#     plot_complexity_fig,
#     plot_dirtyFDF_ax,
#     plot_Ipqu_spectra_fig,
# )

# from rm_lite.tools_1d.rmsynth import saveOutput
# from rm_lite.utils.fitting import (
#     MAD,
#     create_frac_spectra,
#     nanmedian,
#     poly5,
#     progress,
#     toscalar,
# )
# from rm_lite.utils.synthesis import (
#     calc_parabola_vertex,
#     extrap,
#     fit_rmsf,
#     measure_qu_complexity,
# )

# if sys.version_info.major == 2:
#     print("RM-tools will no longer run with Python 2! Please use Python 3.")
#     exit()


# C = 2.997924538e8  # Speed of light [m/s]


# # -----------------------------------------------------------------------------#
# def rotation_integral_limit(freq, phi):
#     """Calculates the analytic solution to the channel polarization integral
#     at one limit frequency.

#     Parameters
#     ----------
#     freq: float
#           frequency in Hz

#     phi: float
#           Faraday depth (rad/m^2)

#     Returns
#     -------
#     intergral_lim: complex
#                    integral limit

#     """
#     funct1 = freq * np.exp(2.0j * phi * ((C / freq) ** 2))
#     funct2 = C * np.sqrt((np.abs(phi) * np.pi))
#     funct3 = -1.0j + np.sign(phi)
#     funct4 = sp.special.erf(np.sqrt(np.abs(phi)) * (C / freq) * (-1.0j + np.sign(phi)))

#     intergral_lim = funct1 + (funct2 * funct3 * funct4)

#     return intergral_lim


# def rotation_operator(channel_width, channel_center, phi):
#     """Rotation operator for channel with top-hat-in-frequency sensitivity.
#     Computes the net effect on a polarization vector for a channel of given
#     center frequnecy and bandwidth, for a given RM.

#     Parameters
#     ----------
#     channel_width: float
#                    channel bandwidth in Hz

#     channel_center: float
#                     channel frequency in Hz

#     phi: float
#          channel frequency in hz

#     Returns
#     -------
#     (complex) rotation operator for that channel

#     """
#     b = channel_center + 0.5 * channel_width
#     a = channel_center - 0.5 * channel_width

#     int_a = rotation_integral_limit(a, phi)
#     int_b = rotation_integral_limit(b, phi)

#     return (1 / channel_width) * (int_b - int_a)


# def estimate_channel_bandwidth(freq_arr):
#     """Estimates the bandwidth per channel given the spacing between channel
#     centers. Only looks at the first 2 channels.

#     Parameters
#     ----------
#     freq_arr: array-like
#                 array of channel centers

#     Returns
#     -------
#     ban: float
#          separation between first two channel centers

#     """
#     ban = freq_arr[1] - freq_arr[0]
#     return ban


# def adjoint_theory(adjoint_vars, dQUArr, show_progress=False, log=print):
#     """Calculates the theoretical sensitivity and noise for the adjoint method

#     Parameters
#     ----------
#     adjoint_vars: list
#                   list like object containing organized as
#                   [widths_Hz, freqArr_Hz, phiArr_radm2, K, weightArr]

#     dQUArr: array like
#             array containing the error in Stokes Q, and U

#     show_progress: Boolean
#                    If set to True, shows progress, Default is False

#     log: function
#          logging function, default is print

#     Returns
#     -------
#     adjoint_info: list
#                   list containing phiArr, and the theoretical noise and
#                   sensitivity organized as
#                   [phiArr_radm2, adjoint_sens, adjoint_noise]

#     """

#     widths_Hz, freqArr_Hz, phiArr_radm2, K, weightArr = adjoint_vars
#     adjoint_noise = np.ones(len(phiArr_radm2))
#     adjoint_sens = np.ones(len(phiArr_radm2))

#     nPhi = len(phiArr_radm2)
#     if show_progress:
#         log("Calculating Theoretical Sensitivity & Noise")
#         progress(40, 0)
#     for i in range(nPhi):
#         if show_progress:
#             progress(40, ((i + 1) * 100.0 / nPhi))

#         r_i = rotation_operator(widths_Hz, freqArr_Hz, phiArr_radm2[i])
#         adjoint_noise2 = (
#             np.sum((weightArr * dQUArr) ** 2 * np.abs(r_i) ** 2)
#             / np.sum(weightArr) ** 2
#         )  # equation 34
#         adjoint_noise[i] = np.sqrt(adjoint_noise2)

#         adjoint_sens[i] = K * np.sum(weightArr * (np.abs(r_i) ** 2))

#     adjoint_info = [phiArr_radm2, adjoint_sens, adjoint_noise]
#     return adjoint_info


# def plot_adjoint_info(mylist, units="Jy/beam"):
#     """plots theoretical noise, sensitivity"""
#     fig, ax = plt.subplots(2, dpi=100, figsize=(12, 8))
#     fig.subplots_adjust(wspace=0.4, hspace=0.4)

#     [phiArr_radm2, adjoint_sens, adjoint_noise] = mylist

#     ax[1].plot(
#         phiArr_radm2,
#         adjoint_sens / adjoint_noise * np.max(adjoint_noise),
#     )
#     ax[1].set_xlabel(r"$\phi$ (rad m$^{-2}$)")
#     ax[1].set_ylabel("S:N multiplier")
#     ax[1].set_title("Theoretical S:N after bandwidth depolarization")
#     # plot 2
#     ax[0].plot(
#         phiArr_radm2,
#         adjoint_sens,
#     )
#     ax[0].set_xlabel(r"$\phi$ (rad m$^{-2}$)")
#     ax[0].set_ylabel("Sensitivity")
#     ax[0].set_title("Theoretical Sensitivity after bandwidth depolarization")
#     return


# # -----------------------------------------------------------------------------#


# def analytical_chan_pol(f, ban, phi, xi_knot=0, p=1):
#     """Calculates the average analytic solution to the channel polarization
#     integral per channel

#     Based on equation 13 of Schnitzeler & Lee (2015)

#     Parameters
#     ----------
#     f: float
#        channel center frequency in Hz

#     ban: float
#          channel bandwidth in Hz

#     phi: float
#          Faraday depth value in rad/m^2

#     xi_knot: float
#              initial polarization angle in radians

#     p: float
#        polarzied intensity

#     Returns
#     -------
#     avg_p_tilde: complex
#                  the average complex polarization, for the bandwidth,
#                  real is Q, imaginary is U

#     """
#     a = f - (ban / 2)
#     b = f + (ban / 2)  # integral start and stop values

#     ya = rotation_integral_limit(
#         a,
#         phi,
#     )
#     yb = rotation_integral_limit(
#         b,
#         phi,
#     )  # check orig for xi_knot

#     i = p * (yb - ya)
#     avg_p_tilde = i / ban

#     return avg_p_tilde


# def bwdepol_simulation(peak_rm, freqArr_Hz, widths_Hz):
#     """Farday thin simulated source of the same RM as the measured source,
#     with unit intensity

#     Parameters
#     ----------
#     peak_rm: float
#              peak in Faraday depth value (in rad/m^2) for sim

#     freqArr_hz: array like
#                 frequency array (in Hz)

#     width_Hz: float
#               channel width in Hz

#     Returns
#     -------
#     data:
#          Dirty FDF for the simulated data formatted as a list of arrays
#          [freq_Hz, q, u,  dq, du]

#     """
#     if widths_Hz == None:
#         widths_Hz = estimate_channel_bandwidth(freqArr_Hz)

#     p_tilde = analytical_chan_pol(freqArr_Hz, widths_Hz, peak_rm)
#     size_f = len(freqArr_Hz)
#     dq = np.ones(size_f)
#     du = np.ones(size_f)

#     # format  = [freq_Hz, q, u,  dq, du]
#     data = [freqArr_Hz, np.real(p_tilde), np.imag(p_tilde), dq, du]
#     return data


# # -----------------------------------------------------------------------------#
# def bwdepol_tweakAxFormat(
#     ax,
#     pad=10,
#     loc="upper right",
#     linewidth=1,
#     ncol=1,
#     bbox_to_anchor=(1.00, 1.00),
#     showLeg=True,
# ):
#     """Tweaks some default plotting parameters for the RMSF, returns ax"""
#     # Axis/tic formatting
#     ax.tick_params(pad=pad)
#     for line in ax.get_xticklines() + ax.get_yticklines():
#         line.set_markeredgewidth(linewidth)

#     # Legend formatting
#     if showLeg:
#         leg = ax.legend(
#             numpoints=1,
#             loc=loc,
#             shadow=False,
#             borderaxespad=0.3,
#             ncol=ncol,
#             bbox_to_anchor=bbox_to_anchor,
#         )
#         for t in leg.get_texts():
#             t.set_fontsize("small")
#         leg.get_frame().set_linewidth(0.5)
#         leg.get_frame().set_alpha(0.5)

#     return ax


# def gauss(p, peak_rm):
#     """Return a function to evaluate a Gaussian with parameters
#     off set by peak_rm

#     Parameters
#     ----------
#     p: list
#        parameters for Gaussian, p = [ampplitude, mean, FWHM]

#     peak_rm: float
#              peak in Faraday depth (in rad/m^2), used to center Gaussian

#     Returns
#     -------
#     rfun: function
#           Gaussian with specified parameters, off set my peak_rm

#     """

#     a, b, w = p
#     gfactor = 2.0 * m.sqrt(2.0 * m.log(2.0))
#     s = w / gfactor

#     def rfunc(x):
#         y = a * np.exp(-((x - b - peak_rm) ** 2.0) / (2.0 * s**2.0))
#         return y

#     return rfunc


# def bwdepol_plot_RMSF_ax(
#     ax,
#     phiArr,
#     RMSFArr,
#     peak_rm,
#     fwhmRMSF=None,
#     axisYright=False,
#     axisXtop=False,
#     doTitle=False,
# ):
#     """Modified for bwdepol, Plots each ax for the RMSF plotting"""

#     # Set the axis positions
#     if axisYright:
#         ax.yaxis.tick_right()
#         ax.yaxis.set_label_position("right")
#     if axisXtop:
#         ax.xaxis.tick_top()
#         ax.xaxis.set_label_position("top")

#     # Plot the RMSF
#     ax.step(phiArr, RMSFArr.real, where="mid", color="tab:blue", lw=0.5, label="Real")
#     ax.step(
#         phiArr, RMSFArr.imag, where="mid", color="tab:red", lw=0.5, label="Imaginary"
#     )
#     ax.step(phiArr, np.abs(RMSFArr), where="mid", color="k", lw=1.0, label="PI")
#     ax.axhline(0, color="grey")
#     if doTitle:
#         ax.text(0.05, 0.84, "RMSF", transform=ax.transAxes)

#     # Plot the Gaussian fit
#     if fwhmRMSF is not None:
#         yGauss = np.max(np.abs(RMSFArr)) * gauss([1.0, 0.0, fwhmRMSF], peak_rm)(phiArr)
#         ax.plot(
#             phiArr,
#             yGauss,
#             color="g",
#             marker="None",
#             mfc="w",
#             mec="g",
#             ms=10,
#             label="Gaussian",
#             lw=2.0,
#             ls="--",
#         )

#     # Scaling
#     ax.yaxis.set_major_locator(MaxNLocator(4))
#     ax.xaxis.set_major_locator(MaxNLocator(4))
#     xRange = np.nanmax(phiArr) - np.nanmin(phiArr)
#     ax.set_xlim(np.nanmin(phiArr) - xRange * 0.01, np.nanmax(phiArr) + xRange * 0.01)
#     ax.set_ylabel("Normalised Units")
#     ax.set_xlabel(r"$\phi$ rad m$^{-2}$")
#     ax.axhline(0, color="grey")

#     # Format tweaks
#     ax = bwdepol_tweakAxFormat(ax)
#     ax.autoscale_view(True, True, True)


# def bwdepol_plot_rmsf_fdf_fig(
#     phiArr,
#     FDF,
#     phi2Arr,
#     RMSFArr,
#     peak_rm,
#     fwhmRMSF=None,
#     gaussParm=[],
#     vLine=None,
#     fig=None,
#     units="flux units",
# ):
#     """Modified for bwdepol, Plot the RMSF and FDF on a single figure"""

#     # Default to a pyplot figure
#     if fig == None:
#         fig = plt.figure(figsize=(12.0, 8))
#     # Plot the RMSF
#     ax1 = fig.add_subplot(211)
#     bwdepol_plot_RMSF_ax(
#         ax=ax1,
#         phiArr=phi2Arr,
#         RMSFArr=RMSFArr,
#         peak_rm=peak_rm,
#         fwhmRMSF=fwhmRMSF,
#         doTitle=True,
#     )
#     [label.set_visible(False) for label in ax1.get_xticklabels()]
#     ax1.set_xlabel("")

#     ax2 = fig.add_subplot(212, sharex=ax1)
#     plot_dirtyFDF_ax(
#         ax=ax2,
#         phiArr=phiArr,
#         FDFArr=FDF,
#         gaussParm=gaussParm,
#         vLine=vLine,
#         doTitle=True,
#         units=units,
#     )
#     return fig


# # -----------------------------------------------------------------------------#
# # modified for adjoint
# def bwdepol_get_rmsf_planes(
#     freqArr_Hz,
#     widths_Hz,
#     phiArr_radm2,
#     peak_rm,
#     weightArr=None,
#     mskArr=None,
#     lam0Sq_m2=None,
#     double=True,
#     fitRMSF=True,
#     fitRMSFreal=False,
#     nBits=64,
#     verbose=False,
#     log=print,
# ):
#     """Calculate the Rotation Measure Spread Function from inputs. This version
#     returns a cube (1, 2 or 3D) of RMSF spectra based on the shape of a
#     boolean mask array, where flagged data are True and unflagged data False.
#     If only whole planes (wavelength channels) are flagged then the RMSF is the
#     same for all pixels and the calculation is done once and replicated to the
#     dimensions of the mask. If some isolated voxels are flagged then the RMSF
#     is calculated by looping through each wavelength plane, which can take some
#     time. By default the routine returns the analytical width of the RMSF main
#     lobe but can also use MPFIT to fit a Gaussian.

#     This has been modified from the convientual RMtools_1D version for bwdepol


#     Parameters
#     ----------
#     freqArr_Hz      ... vector of frequency values
#     phiArr_radm2    ... vector of trial Faraday depth values
#     weightArr       ... vector of weights, default [None] is no weighting
#     maskArr         ... cube of mask values used to shape return cube [None]
#     lam0Sq_m2       ... force a reference lambda^2 value (def=calculate) [None]
#     double          ... pad the Faraday depth to double-size [True]
#     fitRMSF         ... fit the main lobe of the RMSF with a Gaussian [False]
#     fitRMSFreal     ... fit RMSF.real, rather than abs(RMSF) [False]
#     nBits           ... precision of data arrays [64]
#     verbose         ... print feedback during calculation [False]
#     log             ... function to be used to output messages [print]


#     Returns
#     -------
#     RMSFcube: array
#               a cube (1, 2 or 3D) of RMSF spectra based on the shape of a
#               boolean mask array

#     phi2Arr: array
#              array of the Faraday Depth (in rad/m^2)

#     fwhmRMSFArr: array
#                  fwhm of the RMSF

#     statArr: array

#     """

#     # Default data types
#     dtFloat = "float" + str(nBits)
#     dtComplex = "complex" + str(2 * nBits)

#     # For cleaning the RMSF should extend by 1/2 on each side in phi-space
#     if double:
#         nPhi = phiArr_radm2.shape[0]
#         nExt = np.ceil(nPhi / 2.0)
#         resampIndxArr = np.arange(2.0 * nExt + nPhi) - nExt
#         phi2Arr = extrap(resampIndxArr, np.arange(nPhi, dtype="int"), phiArr_radm2)
#     else:
#         phi2Arr = phiArr_radm2

#     # Set the weight array
#     if weightArr is None:
#         weightArr = np.ones(freqArr_Hz.shape, dtype=dtFloat)
#     weightArr = np.where(np.isnan(weightArr), 0.0, weightArr)

#     # Set the mask array (default to 1D, no masked channels)
#     if mskArr is None:
#         mskArr = np.zeros_like(freqArr_Hz, dtype="bool")
#         nDims = 1
#     else:
#         mskArr = mskArr.astype("bool")
#         nDims = len(mskArr.shape)

#     # Sanity checks on array sizes
#     if not weightArr.shape == freqArr_Hz.shape:
#         log("Err: wavelength^2 and weight arrays must be the same shape.")
#         return None, None, None, None
#     if not nDims <= 3:
#         log("Err: mask dimensions must be <= 3.")
#         return None, None, None, None
#     if not mskArr.shape[0] == freqArr_Hz.shape[0]:
#         log("Err: mask depth does not match lambda^2 vector (%d vs %d).", end=" ")
#         (mskArr.shape[0], freqArr_Hz.shape[-1])
#         log("     Check that the mask is in [z, y, x] order.")
#         return None, None, None, None

#     # Reshape the mask array to 3 dimensions
#     if nDims == 1:
#         mskArr = np.reshape(mskArr, (mskArr.shape[0], 1, 1))
#     elif nDims == 2:
#         mskArr = np.reshape(mskArr, (mskArr.shape[0], mskArr.shape[1], 1))

#     # Initialise the complex RM Spread Function cube
#     nX = mskArr.shape[-1]
#     nY = mskArr.shape[-2]
#     nPix = nX * nY
#     nPhi = phi2Arr.shape[0]
#     RMSFcube = np.ones((nPhi, nY, nX), dtype=dtComplex)

#     # If full planes are flagged then set corresponding weights to zero
#     xySum = np.sum(np.sum(mskArr, axis=1), axis=1)
#     mskPlanes = np.where(xySum == nPix, 0, 1)
#     weightArr *= mskPlanes

#     # Check for isolated clumps of flags (# flags in a plane not 0 or nPix)
#     flagTotals = np.unique(xySum).tolist()
#     try:
#         flagTotals.remove(0)
#     except Exception:
#         pass
#     try:
#         flagTotals.remove(nPix)
#     except Exception:
#         pass

#     lambdaSqArr_m2 = np.power(C / freqArr_Hz, 2.0)
#     # Calculate the analytical FWHM width of the main lobe
#     fwhmRMSF = (
#         2.0 * m.sqrt(3.0) / (np.nanmax(lambdaSqArr_m2) - np.nanmin(lambdaSqArr_m2))
#     )

#     # Create simulated data set with simRM = peakRM
#     RMSF_data = bwdepol_simulation(peak_rm, freqArr_Hz, widths_Hz)
#     # RMSFArr = fdf from bwdepol_simulation
#     RMSFArr, _, _ = do_adjoint_rmsynth_planes(
#         freqArr_Hz,
#         RMSF_data[1],
#         RMSF_data[2],
#         phiArr_radm2,
#         widths_Hz=widths_Hz,
#         weightArr=weightArr,
#         lam0Sq_m2=lam0Sq_m2,
#         verbose=verbose,
#         log=print,
#     )

#     # Fit the RMSF main lobe
#     fitStatus = -1
#     if fitRMSF:
#         if verbose:
#             log("Fitting Gaussian to the main lobe.")
#         mp = fit_rmsf(phi2Arr, np.abs(RMSFArr) / np.max(np.abs(RMSFArr)))
#         if mp is None or mp.status < 1:
#             pass
#             log("Err: failed to fit the RMSF.")
#             log("     Defaulting to analytical value.")
#         else:
#             fwhmRMSF = mp.params[2]
#             fitStatus = mp.status

#     # Replicate along X and Y axes
#     RMSFcube = np.tile(RMSFArr[:, np.newaxis, np.newaxis], (1, nY, nX))
#     fwhmRMSFArr = np.ones((nY, nX), dtype=dtFloat) * fwhmRMSF
#     statArr = np.ones((nY, nX), dtype="int") * fitStatus

#     # Remove redundant dimensions
#     fwhmRMSFArr = np.squeeze(fwhmRMSFArr)
#     statArr = np.squeeze(statArr)
#     RMSFcube = RMSFcube.reshape(-1)

#     return RMSFcube, phi2Arr, fwhmRMSFArr, statArr


# # -----------------------------------------------------------------------------#
# def bwdepol_measure_FDF_parms(
#     FDF,
#     phiArr,
#     fwhmRMSF,
#     adjoint_sens,
#     adjoint_noise,
#     dFDF=None,
#     lamSqArr_m2=None,
#     lam0Sq=None,
#     snrDoBiasCorrect=5.0,
# ):
#     """
#     Measure standard parameters from a complex Faraday Dispersion Function.
#     Currently this function assumes that the noise levels in the Stokes Q
#     and U spectra are the same.
#     Returns a dictionary containing measured parameters.

#     This has been modified from the convientual RMtools_1D version for bwdepol
#     """

#     # Determine the peak channel in the FDF, its amplitude and index
#     absFDF = np.abs(FDF)
#     rm_fdf = (
#         absFDF / adjoint_noise
#     )  # RM spectrum in S:N units (normalized by RM-dependent noise)
#     amp_fdf = (
#         absFDF / adjoint_sens
#     )  # RM spectrum normalized by (RM-dependent) sensitivity
#     indxPeakPIchan = np.nanargmax(rm_fdf[1:-1]) + 1  # Masks out the edge channels

#     # new theoretical dFDF correction for adjoint method
#     # This is noise in the adjoint-spectrum.
#     dFDF = adjoint_noise[indxPeakPIchan]

#     # This is the error in the amplitude (accounting for re-normalization)
#     dampPeakPI = dFDF / adjoint_sens[indxPeakPIchan]

#     # Measure the RMS noise in the spectrum after masking the peak
#     # changed all absFDF to rm_fdf
#     # Since this is normalized by theoretical noise, it's effectively testing
#     # the noise relative to the theoretical noise.
#     dPhi = np.nanmin(np.diff(phiArr))
#     fwhmRMSF_chan = np.ceil(fwhmRMSF / dPhi)
#     iL = int(max(0, indxPeakPIchan - fwhmRMSF_chan * 2))
#     iR = int(min(len(absFDF), indxPeakPIchan + fwhmRMSF_chan * 2))
#     absFDFmsked = rm_fdf.copy()
#     absFDFmsked[iL:iR] = np.nan
#     absFDFmsked = absFDFmsked[np.where(absFDFmsked == absFDFmsked)]
#     if float(len(absFDFmsked)) / len(absFDF) < 0.3:
#         dFDFcorMAD = MAD(rm_fdf)
#     else:
#         dFDFcorMAD = MAD(absFDFmsked)

#     # The noise is re-normalized by the predicted noise at the peak RM.
#     dFDFcorMAD = dFDFcorMAD * adjoint_noise[indxPeakPIchan]

#     nChansGood = np.sum(np.where(lamSqArr_m2 == lamSqArr_m2, 1.0, 0.0))
#     varLamSqArr_m2 = (
#         np.sum(lamSqArr_m2**2.0) - np.sum(lamSqArr_m2) ** 2.0 / nChansGood
#     ) / (nChansGood - 1)

#     # Determine the peak in the FDF, its amplitude and Phi using a
#     # 3-point parabolic interpolation
#     phiPeakPIfit = None
#     dPhiPeakPIfit = None
#     ampPeakPIfit = None
#     snrPIfit = None
#     ampPeakPIfitEff = None
#     indxPeakPIfit = None
#     peakFDFimagFit = None
#     peakFDFrealFit = None
#     polAngleFit_deg = None
#     dPolAngleFit_deg = None
#     polAngle0Fit_deg = None
#     dPolAngle0Fit_deg = None

#     # Only do the 3-point fit if peak is 1-channel from either edge
#     if indxPeakPIchan > 0 and indxPeakPIchan < len(FDF) - 1:
#         phiPeakPIfit, ampPeakPIfit = calc_parabola_vertex(
#             phiArr[indxPeakPIchan - 1],
#             amp_fdf[indxPeakPIchan - 1],
#             phiArr[indxPeakPIchan],
#             amp_fdf[indxPeakPIchan],
#             phiArr[indxPeakPIchan + 1],
#             amp_fdf[indxPeakPIchan + 1],
#         )

#         snrPIfit = ampPeakPIfit * adjoint_sens[indxPeakPIchan] / dFDF

#         # Error on fitted Faraday depth (RM) is same as channel
#         # but using fitted PI
#         dPhiPeakPIfit = fwhmRMSF / (2.0 * snrPIfit)

#         # Correct the peak for polarisation bias (POSSUM report 11)
#         ampPeakPIfitEff = ampPeakPIfit
#         if snrPIfit >= snrDoBiasCorrect:
#             ampPeakPIfitEff = np.sqrt(ampPeakPIfit**2.0 - 2.3 * dampPeakPI**2.0)

#         # Calculate the polarisation angle from the fitted peak
#         # Uncertainty from Eqn A.12 in Brentjens & De Bruyn 2005
#         indxPeakPIfit = np.interp(
#             phiPeakPIfit, phiArr, np.arange(phiArr.shape[-1], dtype="f4")
#         )
#         peakFDFimagFit = np.interp(phiPeakPIfit, phiArr, FDF.imag)
#         peakFDFrealFit = np.interp(phiPeakPIfit, phiArr, FDF.real)
#         polAngleFit_deg = (
#             0.5 * np.degrees(np.arctan2(peakFDFimagFit, peakFDFrealFit)) % 180
#         )
#         dPolAngleFit_deg = np.degrees(1 / (2.0 * snrPIfit))

#         # Calculate the derotated polarisation angle and uncertainty
#         # Uncertainty from Eqn A.20 in Brentjens & De Bruyn 2005
#         polAngle0Fit_deg = (
#             np.degrees(np.radians(polAngleFit_deg) - phiPeakPIfit * lam0Sq)
#         ) % 180
#         dPolAngle0Fit_rad = np.sqrt(
#             nChansGood
#             / (4.0 * (nChansGood - 2.0) * snrPIfit**2.0)
#             * ((nChansGood - 1) / nChansGood + lam0Sq**2.0 / varLamSqArr_m2)
#         )
#         dPolAngle0Fit_deg = np.degrees(dPolAngle0Fit_rad)

#     # Store the measurements in a dictionary and return
#     mDict = {
#         "dFDFcorMAD": toscalar(dFDFcorMAD),
#         "phiPeakPIfit_rm2": toscalar(phiPeakPIfit),
#         "dPhiPeakPIfit_rm2": toscalar(dPhiPeakPIfit),
#         "ampPeakPIfit": toscalar(ampPeakPIfit),
#         "ampPeakPIfitEff": toscalar(ampPeakPIfitEff),
#         "dAmpPeakPIfit": toscalar(dampPeakPI),
#         "snrPIfit": toscalar(snrPIfit),
#         "indxPeakPIfit": toscalar(indxPeakPIfit),
#         "peakFDFimagFit": toscalar(peakFDFimagFit),
#         "peakFDFrealFit": toscalar(peakFDFrealFit),
#         "polAngleFit_deg": toscalar(polAngleFit_deg),
#         "dPolAngleFit_deg": toscalar(dPolAngleFit_deg),
#         "polAngle0Fit_deg": toscalar(polAngle0Fit_deg),
#         "dPolAngle0Fit_deg": toscalar(dPolAngle0Fit_deg),
#     }

#     return mDict


# # -----------------------------------------------------------------------------#
# def do_adjoint_rmsynth_planes(
#     freqArr_Hz,
#     dataQ,
#     dataU,
#     phiArr_radm2,
#     widths_Hz=None,
#     weightArr=None,
#     lam0Sq_m2=None,
#     nBits=64,
#     verbose=False,
#     log=print,
# ):
#     """Perform RM-synthesis on Stokes Q and U cubes (1,2 or 3D). This version
#     of the routine loops through spectral planes and is faster than the pixel-
#     by-pixel code. This version also correctly deals with isolated clumps of
#     NaN-flagged voxels within the data-cube (unlikely in interferometric cubes,
#     but possible in single-dish cubes). Input data must be in standard python
#     [z,y,x] order, where z is the frequency axis in ascending order.

#     This has been modified from the convientual RMtools_1D version for bwdepol

#     Parameters
#     ----------
#     dataQ           ... 1, 2 or 3D Stokes Q data array
#     dataU           ... 1, 2 or 3D Stokes U data array
#     lambdaSqArr_m2  ... vector of wavelength^2 values (ascending freq order)
#     phiArr_radm2    ... vector of trial Faraday depth values
#     weightArr       ... vector of weights, default [None] is Uniform (all 1s)
#     nBits           ... precision of data arrays [32]
#     verbose         ... print feedback during calculation [False]
#     log             ... function to be used to output messages [print]

#     Returns
#     -------
#     FDFcube: array
#              Faraday Dispersion Function (FDF)

#     lam0Sq_m2: array
#                lam0Sq_m2 is the weighted mean of lambda^2 distribution
#                (B&dB Eqn. 32)

#     adjoint_vars: list
#                   information to generate theoretical noise, sensitivity
#                   adjoint_vars = [widths_Hz, freqArr_Hz, phiArr_radm2, K,
#                                   weightArr]

#     """

#     # Default data types
#     dtFloat = "float" + str(nBits)
#     dtComplex = "complex" + str(2 * nBits)

#     lambdaSqArr_m2 = np.power(C / freqArr_Hz, 2.0)
#     # Set the weight array
#     if weightArr is None:
#         weightArr = np.ones(lambdaSqArr_m2.shape, dtype=dtFloat)
#     weightArr = np.where(np.isnan(weightArr), 0.0, weightArr)

#     # Sanity check on array sizes
#     if not weightArr.shape == lambdaSqArr_m2.shape:
#         log("Err: Lambda^2 and weight arrays must be the same shape.")
#         return None, None
#     if not dataQ.shape == dataU.shape:
#         log("Err: Stokes Q and U data arrays must be the same shape.")
#         return None, None
#     nDims = len(dataQ.shape)
#     if not nDims <= 3:
#         log("Err: data dimensions must be <= 3.")
#         return None, None
#     if not dataQ.shape[0] == lambdaSqArr_m2.shape[0]:
#         log(
#             "Err: Data depth does not match lambda^2 vector ({} vs {}).".format(
#                 dataQ.shape[0], lambdaSqArr_m2.shape[0]
#             ),
#             end=" ",
#         )
#         log("     Check that data is in [z, y, x] order.")
#         return None, None

#     # Reshape the data arrays to 3 dimensions
#     if nDims == 1:
#         dataQ = np.reshape(dataQ, (dataQ.shape[0], 1, 1))
#         dataU = np.reshape(dataU, (dataU.shape[0], 1, 1))
#     elif nDims == 2:
#         dataQ = np.reshape(dataQ, (dataQ.shape[0], dataQ.shape[1], 1))
#         dataU = np.reshape(dataU, (dataU.shape[0], dataU.shape[1], 1))

#     # Create a complex polarised cube, B&dB Eqns. (8) and (14)
#     # Array has dimensions [nFreq, nY, nX]
#     pCube = (dataQ + 1j * dataU) * weightArr[:, np.newaxis, np.newaxis]

#     # Check for NaNs (flagged data) in the cube & set to zero
#     mskCube = np.isnan(pCube)
#     pCube = np.nan_to_num(pCube)

#     # If full planes are flagged then set corresponding weights to zero
#     mskPlanes = np.sum(np.sum(~mskCube, axis=1), axis=1)
#     mskPlanes = np.where(mskPlanes == 0, 0, 1)
#     weightArr *= mskPlanes

#     # Initialise the complex Faraday Dispersion Function cube
#     nX = dataQ.shape[-1]
#     nY = dataQ.shape[-2]
#     nPhi = phiArr_radm2.shape[0]
#     FDFcube = np.zeros((nPhi, nY, nX), dtype=dtComplex)

#     # lam0Sq_m2 is the weighted mean of lambda^2 distribution (B&dB Eqn. 32)
#     # Calculate a global lam0Sq_m2 value, ignoring isolated flagged voxels
#     K = 1.0 / np.sum(weightArr)
#     if lam0Sq_m2 is None:
#         lam0Sq_m2 = K * np.sum(weightArr * lambdaSqArr_m2)

#     # The K value used to scale each FDF spectrum must take into account
#     # flagged voxels data in the datacube and can be position dependent
#     weightCube = np.invert(mskCube) * weightArr[:, np.newaxis, np.newaxis]
#     with np.errstate(divide="ignore", invalid="ignore"):
#         KArr = np.true_divide(1.0, np.sum(weightCube, axis=0))
#         KArr[KArr == np.inf] = 0
#         KArr = np.nan_to_num(KArr)

#     # Do the RM-synthesis on each plane
#     if verbose:
#         log("Running RM-synthesis by channel.")
#         progress(40, 0)

#     # calculate channel widths if necessary
#     if widths_Hz == None:
#         widths_Hz = estimate_channel_bandwidth(freqArr_Hz)
#     for i in range(nPhi):
#         if verbose:
#             progress(40, ((i + 1) * 100.0 / nPhi))
#         cor = np.exp(2j * phiArr_radm2[i] * lam0Sq_m2)
#         r_i = rotation_operator(widths_Hz, freqArr_Hz, phiArr_radm2[i])[
#             :, np.newaxis, np.newaxis
#         ]
#         arg0 = pCube * cor * np.conj(r_i)
#         arg = arg0
#         FDFcube[i, :, :] = KArr * np.sum(arg, axis=0)

#     # information to generate theoretical noise, sensitivity
#     adjoint_vars = [widths_Hz, freqArr_Hz, phiArr_radm2, K, weightArr]

#     # Remove redundant dimensions in the FDF array
#     FDFcube = np.squeeze(FDFcube)
#     return FDFcube, lam0Sq_m2, adjoint_vars


# # -----------------------------------------------------------------------------#
# def run_adjoint_rmsynth(
#     data,
#     polyOrd=3,
#     phiMax_radm2=None,
#     dPhi_radm2=None,
#     nSamples=10.0,
#     weightType="variance",
#     fitRMSF=True,
#     noStokesI=False,
#     phiNoise_radm2=1e6,
#     nBits=64,
#     showPlots=False,
#     debug=False,
#     verbose=False,
#     log=print,
#     units="Jy/beam",
# ):
#     """Run bwdepol RM synthesis on 1D data.

#     Args:
#         data (list): Contains frequency and polarization data as either:
#             [freq_Hz, I, Q, U, dI, dQ, dU]
#                 freq_Hz (array_like): Frequency of each channel in Hz.
#                 I (array_like): Stokes I intensity in each channel.
#                 Q (array_like): Stokes Q intensity in each channel.
#                 U (array_like): Stokes U intensity in each channel.
#                 dI (array_like): Error in Stokes I intensity in each channel.
#                 dQ (array_like): Error in Stokes Q intensity in each channel.
#                 dU (array_like): Error in Stokes U intensity in each channel.
#             or
#             [freq_Hz, q, u,  dq, du]
#                 freq_Hz (array_like): Frequency of each channel in Hz.
#                 q (array_like): Fractional Stokes Q intensity (Q/I) in each channel.
#                 u (array_like): Fractional Stokes U intensity (U/I) in each channel.
#                 dq (array_like): Error in fractional Stokes Q intensity in each channel.
#                 du (array_like): Error in fractional Stokes U intensity in each channel.

#     Kwargs:
#         polyOrd (int): Order of polynomial to fit to Stokes I spectrum.
#         phiMax_radm2 (float): Maximum absolute Faraday depth (rad/m^2).
#         dPhi_radm2 (float): Faraday depth channel size (rad/m^2).
#         nSamples (float): Number of samples across the RMSF.
#         weightType (str): Can be "variance" or "uniform"
#             "variance" -- Weight by uncertainty in Q and U.
#             "uniform" -- Weight uniformly (i.e. with 1s)
#         fitRMSF (bool): Fit a Gaussian to the RMSF?
#         noStokesI (bool: Is Stokes I data provided?
#         phiNoise_radm2 (float): ????
#         nBits (int): Precision of floating point numbers.
#         showPlots (bool): Show plots?
#         debug (bool): Turn on debugging messages & plots?
#         verbose (bool): Verbosity.
#         log (function): Which logging function to use.
#         units (str): Units of data.

#     Returns:
#         mDict (dict): Summary of RM synthesis results.
#         aDict (dict): Data output by RM synthesis.

#     """

#     # Default data types
#     dtFloat = "float" + str(nBits)
#     #    dtComplex = "complex" + str(2*nBits)

#     # freq_Hz, I, Q, U, dI, dQ, dU
#     if data.shape[0] == 7:
#         if verbose:
#             log("> Seven columns found, trying [freq_Hz, I, Q, U, dI, dQ, dU]", end=" ")
#         (freqArr_Hz, IArr, QArr, UArr, dIArr, dQArr, dUArr) = data
#         widths_Hz = None
#     elif data.shape[0] == 8:
#         if verbose:
#             log(
#                 "> Eight columns found, trying [freq_Hz, widths_Hz, I, Q, U, dI, dQ, dU]",
#                 end=" ",
#             )
#         (freqArr_Hz, widths_Hz, IArr, QArr, UArr, dIArr, dQArr, dUArr) = data
#     elif data.shape[0] == 6:
#         if verbose:
#             log(
#                 "> Six columns found, trying [freq_Hz, widths_Hz, Q, U, dQ, dU]",
#                 end=" ",
#             )
#         (freqArr_Hz, widths_Hz, QArr, UArr, dQArr, dUArr) = data
#     elif data.shape[0] == 5:
#         if verbose:
#             log("> Five columns found, trying [freq_Hz, Q, U, dQ, dU]", end=" ")
#         (freqArr_Hz, QArr, UArr, dQArr, dUArr) = data
#         widths_Hz = None
#         noStokesI = True
#     else:
#         log("Failed to read in data, aborting.")
#         if debug:
#             log(traceback.format_exc())
#         sys.exit()
#     if verbose:
#         log("Successfully read in the Stokes spectra.")

#     # If no Stokes I present, create a dummy spectrum = unity
#     if noStokesI:
#         if verbose:
#             log("Warn: no Stokes I data in use.")
#         IArr = np.ones_like(QArr)
#         dIArr = np.zeros_like(QArr)

#     # Convert to GHz for convenience
#     freqArr_GHz = freqArr_Hz / 1e9
#     dQUArr = (dQArr + dUArr) / 2.0

#     # Fit the Stokes I spectrum and create the fractional spectra
#     IModArr, qArr, uArr, dqArr, duArr, fit_result = create_frac_spectra(
#         freqArr=freqArr_GHz,
#         IArr=IArr,
#         QArr=QArr,
#         UArr=UArr,
#         dIArr=dIArr,
#         dQArr=dQArr,
#         dUArr=dUArr,
#         polyOrd=polyOrd,
#         verbose=True,
#         debug=debug,
#     )

#     # Plot the data and the Stokes I model fit
#     if showPlots:
#         if verbose:
#             log("Plotting the input data and spectral index fit.")
#         freqHirArr_Hz = np.linspace(freqArr_Hz[0], freqArr_Hz[-1], 10000)
#         IModHirArr = poly5(fit_result.params)(freqHirArr_Hz / 1e9)
#         specFig = plt.figure(figsize=(12.0, 8))
#         plot_Ipqu_spectra_fig(
#             freqArr_Hz=freqArr_Hz,
#             IArr=IArr,
#             qArr=qArr,
#             uArr=uArr,
#             dIArr=dIArr,
#             dqArr=dqArr,
#             duArr=duArr,
#             freqHirArr_Hz=freqHirArr_Hz,
#             IModArr=IModHirArr,
#             fig=specFig,
#             units=units,
#         )

#         # DEBUG (plot the Q, U and average RMS spectrum)
#         if debug:
#             rmsFig = plt.figure(figsize=(12.0, 8))
#             ax = rmsFig.add_subplot(111)
#             ax.plot(
#                 freqArr_Hz / 1e9,
#                 dQUArr,
#                 marker="o",
#                 color="k",
#                 lw=0.5,
#                 label="rms <QU>",
#             )
#             ax.plot(
#                 freqArr_Hz / 1e9, dQArr, marker="o", color="b", lw=0.5, label="rms Q"
#             )
#             ax.plot(
#                 freqArr_Hz / 1e9, dUArr, marker="o", color="r", lw=0.5, label="rms U"
#             )
#             xRange = (np.nanmax(freqArr_Hz) - np.nanmin(freqArr_Hz)) / 1e9
#             ax.set_xlim(
#                 np.min(freqArr_Hz) / 1e9 - xRange * 0.05,
#                 np.max(freqArr_Hz) / 1e9 + xRange * 0.05,
#             )
#             ax.set_xlabel(r"$\nu$ (GHz)")
#             ax.set_ylabel("RMS " + units)
#             ax.set_title("RMS noise in Stokes Q, U and <Q,U> spectra")

#     # Calculate some wavelength parameters
#     lambdaSqArr_m2 = np.power(C / freqArr_Hz, 2.0)
#     # dFreq_Hz = np.nanmin(np.abs(np.diff(freqArr_Hz)))
#     lambdaSqRange_m2 = np.nanmax(lambdaSqArr_m2) - np.nanmin(lambdaSqArr_m2)
#     # dLambdaSqMin_m2 = np.nanmin(np.abs(np.diff(lambdaSqArr_m2)))
#     dLambdaSqMax_m2 = np.nanmax(np.abs(np.diff(lambdaSqArr_m2)))

#     # Set the Faraday depth range
#     fwhmRMSF_radm2 = 2.0 * m.sqrt(3.0) / lambdaSqRange_m2
#     if dPhi_radm2 is None:
#         dPhi_radm2 = fwhmRMSF_radm2 / nSamples
#     if phiMax_radm2 is None:
#         phiMax_radm2 = m.sqrt(3.0) / dLambdaSqMax_m2
#         phiMax_radm2 = max(phiMax_radm2, 600.0)  # Force the minimum phiMax

#     # Faraday depth sampling. Zero always centred on middle channel
#     nChanRM = int(round(abs((phiMax_radm2 - 0.0) / dPhi_radm2)) * 2.0 + 1.0)
#     startPhi_radm2 = -(nChanRM - 1.0) * dPhi_radm2 / 2.0
#     stopPhi_radm2 = +(nChanRM - 1.0) * dPhi_radm2 / 2.0
#     phiArr_radm2 = np.linspace(startPhi_radm2, stopPhi_radm2, nChanRM)
#     phiArr_radm2 = phiArr_radm2.astype(dtFloat)
#     if verbose:
#         log(
#             "PhiArr = %.2f to %.2f by %.2f (%d chans)."
#             % (phiArr_radm2[0], phiArr_radm2[-1], float(dPhi_radm2), nChanRM)
#         )

#     # Calculate the weighting as 1/sigma^2 or all 1s (uniform)
#     if weightType == "variance":
#         weightArr = 1.0 / np.power(dQUArr, 2.0)
#     else:
#         weightType = "uniform"
#         weightArr = np.ones(freqArr_Hz.shape, dtype=dtFloat)
#     if verbose:
#         log("Weight type is '%s'." % weightType)

#     startTime = time.time()

#     # Perform adjoint RM-synthesis on the spectrum
#     dirtyFDF, lam0Sq_m2, adjoint_vars = do_adjoint_rmsynth_planes(
#         freqArr_Hz=freqArr_Hz,
#         widths_Hz=widths_Hz,
#         dataQ=qArr,
#         dataU=uArr,
#         phiArr_radm2=phiArr_radm2,
#         weightArr=weightArr,
#         nBits=nBits,
#         verbose=verbose,
#         log=log,
#     )

#     # generate adjoint_noise and adjoint__sens
#     adjoint_info = adjoint_theory(adjoint_vars, dQUArr, show_progress=False)
#     phiArr_radm2, adjoint_sens, adjoint_noise = adjoint_info

#     # calculate peak RM
#     absFDF = np.abs(dirtyFDF)
#     rm_fdf = absFDF / adjoint_noise  # used for finding peak in RM
#     indxPeakPIchan = np.nanargmax(rm_fdf[1:-1]) + 1
#     peak_rm = phiArr_radm2[indxPeakPIchan]

#     # Calculate the Rotation Measure Spread Function
#     RMSFArr, phi2Arr_radm2, fwhmRMSFArr, fitStatArr = bwdepol_get_rmsf_planes(
#         freqArr_Hz=freqArr_Hz,
#         widths_Hz=widths_Hz,
#         phiArr_radm2=phiArr_radm2,
#         weightArr=weightArr,
#         mskArr=~np.isfinite(qArr),
#         lam0Sq_m2=lam0Sq_m2,
#         double=True,
#         fitRMSF=fitRMSF,
#         fitRMSFreal=False,
#         nBits=nBits,
#         verbose=verbose,
#         log=log,
#         peak_rm=peak_rm,
#     )
#     fwhmRMSF = float(fwhmRMSFArr)
#     endTime = time.time()
#     cputime = endTime - startTime
#     if verbose:
#         log("> RM-synthesis completed in %.2f seconds." % cputime)

#     # Determine the Stokes I value at lam0Sq_m2 from the Stokes I model
#     # Multiply the dirty FDF by Ifreq0 to recover the PI
#     freq0_Hz = C / m.sqrt(lam0Sq_m2)
#     Ifreq0 = poly5(fit_result.params)(freq0_Hz / 1e9)
#     dirtyFDF *= Ifreq0  # FDF is in fracpol units initially, convert back to flux

#     # Calculate the theoretical noise in the FDF !!
#     # Old formula only works for wariance weights!
#     weightArr = np.where(np.isnan(weightArr), 0.0, weightArr)
#     dFDFth = np.sqrt(
#         np.sum(weightArr**2 * np.nan_to_num(dQUArr) ** 2) / (np.sum(weightArr)) ** 2
#     )

#     # Measure the parameters of the dirty FDF
#     # Use the theoretical noise to calculate uncertainties

#     mDict = bwdepol_measure_FDF_parms(
#         FDF=dirtyFDF,
#         phiArr=phiArr_radm2,
#         fwhmRMSF=fwhmRMSF,
#         adjoint_sens=adjoint_sens,
#         adjoint_noise=adjoint_noise,
#         dFDF=dFDFth,
#         lamSqArr_m2=lambdaSqArr_m2,
#         lam0Sq=lam0Sq_m2,
#     )

#     mDict["Ifreq0"] = toscalar(Ifreq0)
#     mDict["polyCoeffs"] = ",".join([str(x) for x in fit_result.params])
#     mDict["IfitStat"] = fit_result.fitStatus
#     mDict["IfitChiSqRed"] = fit_result.chiSqRed
#     mDict["lam0Sq_m2"] = toscalar(lam0Sq_m2)
#     mDict["freq0_Hz"] = toscalar(freq0_Hz)
#     mDict["fwhmRMSF"] = toscalar(fwhmRMSF)
#     mDict["dQU"] = toscalar(nanmedian(dQUArr))
#     # mDict["dFDFth"] = toscalar(dFDFth)
#     mDict["units"] = units

#     if fit_result.fitStatus >= 128:
#         log("WARNING: Stokes I model contains negative values!")
#     elif fit_result.fitStatus >= 64:
#         log("Caution: Stokes I model has low signal-to-noise.")

#     # Add information on nature of channels:
#     good_channels = np.where(np.logical_and(weightArr != 0, np.isfinite(qArr)))[0]
#     mDict["min_freq"] = float(np.min(freqArr_Hz[good_channels]))
#     mDict["max_freq"] = float(np.max(freqArr_Hz[good_channels]))
#     mDict["N_channels"] = good_channels.size
#     if widths_Hz != None:
#         mDict["median_channel_width"] = float(np.median(widths_Hz))
#     else:
#         mDict["median_channel_width"] = float(np.median(np.diff(freqArr_Hz)))

#     # Measure the complexity of the q and u spectra
#     mDict["fracPol"] = mDict["ampPeakPIfit"] / (Ifreq0)
#     mD, pD = measure_qu_complexity(
#         freqArr_Hz=freqArr_Hz,
#         qArr=qArr,
#         uArr=uArr,
#         dqArr=dqArr,
#         duArr=duArr,
#         fracPol=mDict["fracPol"],
#         psi0_deg=mDict["polAngle0Fit_deg"],
#         RM_radm2=mDict["phiPeakPIfit_rm2"],
#     )
#     mDict.update(mD)

#     # Debugging plots for spectral complexity measure
#     if debug:
#         tmpFig = plot_complexity_fig(
#             xArr=pD["xArrQ"],
#             qArr=pD["yArrQ"],
#             dqArr=pD["dyArrQ"],
#             sigmaAddqArr=pD["sigmaAddArrQ"],
#             chiSqRedqArr=pD["chiSqRedArrQ"],
#             probqArr=pD["probArrQ"],
#             uArr=pD["yArrU"],
#             duArr=pD["dyArrU"],
#             sigmaAdduArr=pD["sigmaAddArrU"],
#             chiSqReduArr=pD["chiSqRedArrU"],
#             probuArr=pD["probArrU"],
#             mDict=mDict,
#         )
#         tmpFig.show()

#     # add array dictionary
#     aDict = dict()
#     aDict["phiArr_radm2"] = phiArr_radm2
#     aDict["phi2Arr_radm2"] = phi2Arr_radm2
#     aDict["RMSFArr"] = RMSFArr
#     aDict["freqArr_Hz"] = freqArr_Hz
#     aDict["weightArr"] = weightArr
#     aDict["dirtyFDF"] = dirtyFDF / adjoint_sens

#     if verbose:
#         # Print the results to the screen
#         log()
#         log("-" * 80)
#         log("RESULTS:\n")
#         log("FWHM RMSF = %.4g rad/m^2" % (mDict["fwhmRMSF"]))

#         log(
#             "Pol Angle = %.4g (+/-%.4g) deg"
#             % (mDict["polAngleFit_deg"], mDict["dPolAngleFit_deg"])
#         )
#         log(
#             "Pol Angle 0 = %.4g (+/-%.4g) deg"
#             % (mDict["polAngle0Fit_deg"], mDict["dPolAngle0Fit_deg"])
#         )
#         log(
#             "Peak FD = %.4g (+/-%.4g) rad/m^2"
#             % (mDict["phiPeakPIfit_rm2"], mDict["dPhiPeakPIfit_rm2"])
#         )
#         log("freq0_GHz = %.4g " % (mDict["freq0_Hz"] / 1e9))
#         log("I freq0 = %.4g %s" % (mDict["Ifreq0"], units))
#         log(
#             "Peak PI = %.4g (+/-%.4g) %s"
#             % (mDict["ampPeakPIfit"], mDict["dAmpPeakPIfit"], units)
#         )
#         log("QU Noise = %.4g %s" % (mDict["dQU"], units))
#         log("FDF Noise (theory)   = %.4g %s" % (mDict["dFDFth"], units))
#         log("FDF Noise (Corrected MAD) = %.4g %s" % (mDict["dFDFcorMAD"], units))
#         log("FDF SNR = %.4g " % (mDict["snrPIfit"]))
#         log(
#             "sigma_add(q) = %.4g (+%.4g, -%.4g)"
#             % (mDict["sigmaAddQ"], mDict["dSigmaAddPlusQ"], mDict["dSigmaAddMinusQ"])
#         )
#         log(
#             "sigma_add(u) = %.4g (+%.4g, -%.4g)"
#             % (mDict["sigmaAddU"], mDict["dSigmaAddPlusU"], mDict["dSigmaAddMinusU"])
#         )
#         log()
#         log("-" * 80)

#     # Plot the RM Spread Function and dirty FDF
#     if showPlots:
#         plot_adjoint_info(adjoint_info, units=units)
#         fdfFig = plt.figure(figsize=(12.0, 8))
#         bwdepol_plot_rmsf_fdf_fig(
#             phiArr=phiArr_radm2,
#             FDF=(dirtyFDF / adjoint_sens),
#             phi2Arr=phiArr_radm2,
#             RMSFArr=RMSFArr,
#             peak_rm=peak_rm,
#             fwhmRMSF=fwhmRMSF,
#             vLine=mDict["phiPeakPIfit_rm2"],
#             fig=fdfFig,
#             units=units,
#         )

#     if showPlots or debug:
#         plt.show()

#     return mDict, aDict


# # -----------------------------------------------------------------------------#
# def main():
#     """
#     Start the function to perform bwdepol RM-synthesis if called from the command line.
#     """

#     # Help string to be shown using the -h option
#     descStr = """
#     Run bandwidth-depolarization-corrected RM-synthesis (based on Fine et al 2022)
#     on Stokes I, Q and U spectra (1D) stored in an ASCII file.

#     Behaves similarly to rmsynth1d except that the input file can optionally
#     contain a column with the channel widths in Hz. If this column is not
#     given, the channel widths will be assumed to be uniform and calculated
#     based on the difference between the frequencies of the first two channels.

#     The ASCII file requires one of the following column configurations,
#     depending on whether Stokes I and channel width information are available,
#     in a space separated format:
#     [freq_Hz, I, Q, U, I_err, Q_err, U_err]
#     [freq_Hz, widths_Hz, I, Q, U, I_err, Q_err, U_err]
#     [freq_Hz, Q, U, Q_err, U_err]
#     [freq_Hz, widths_Hz, Q, U, Q_err, U_err]


#     To get outputs, one or more of the following flags must be set: -S, -p, -v.
#     """

#     epilog_text = """
#     Outputs with -S flag:
#     _FDFdirty.dat: Dirty FDF/RM Spectrum [Phi, Q, U]
#     _RMSF.dat: Computed RMSF [Phi, Q, U]
#     _RMsynth.dat: list of derived parameters for RM spectrum
#                 (approximately equivalent to -v flag output)
#     _RMsynth.json: dictionary of derived parameters for RM spectrum
#     _weight.dat: Calculated channel weights [freq_Hz, weight]
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr,
#         epilog=epilog_text,
#         formatter_class=argparse.RawTextHelpFormatter,
#     )
#     parser.add_argument(
#         "dataFile",
#         metavar="dataFile.dat",
#         nargs=1,
#         help="ASCII file containing Stokes spectra & errors.",
#     )
#     parser.add_argument(
#         "-t",
#         dest="fitRMSF",
#         action="store_false",
#         help="fit a Gaussian to the RMSF [True; set flag to disable]",
#     )
#     parser.add_argument(
#         "-l",
#         dest="phiMax_radm2",
#         type=float,
#         default=None,
#         help="absolute max Faraday depth sampled [Auto].",
#     )
#     parser.add_argument(
#         "-d",
#         dest="dPhi_radm2",
#         type=float,
#         default=None,
#         help="width of Faraday depth channel [Auto].\n(overrides -s NSAMPLES flag)",
#     )
#     parser.add_argument(
#         "-s",
#         dest="nSamples",
#         type=float,
#         default=10,
#         help="number of samples across the RMSF lobe [10].",
#     )
#     parser.add_argument(
#         "-w",
#         dest="weightType",
#         default="variance",
#         help="weighting [inverse variance] or 'uniform' (all 1s).",
#     )
#     parser.add_argument(
#         "-o",
#         dest="polyOrd",
#         type=int,
#         default=2,
#         help="polynomial order to fit to I spectrum [2].",
#     )
#     parser.add_argument(
#         "-i",
#         dest="noStokesI",
#         action="store_true",
#         help="ignore the Stokes I spectrum [False].",
#     )
#     parser.add_argument(
#         "-p", dest="showPlots", action="store_true", help="show the plots [False]."
#     )
#     parser.add_argument(
#         "-v", dest="verbose", action="store_true", help="verbose output [False]."
#     )
#     parser.add_argument(
#         "-S", dest="saveOutput", action="store_true", help="save the arrays [False]."
#     )
#     parser.add_argument(
#         "-D",
#         dest="debug",
#         action="store_true",
#         help="turn on debugging messages & plots [False].",
#     )
#     parser.add_argument(
#         "-U",
#         dest="units",
#         type=str,
#         default="Jy/beam",
#         help="Intensity units of the data. [Jy/beam]",
#     )
#     args = parser.parse_args()

#     # Sanity checks
#     if not os.path.exists(args.dataFile[0]):
#         print("File does not exist: '%s'." % args.dataFile[0])
#         sys.exit()
#     (prefixOut,) = os.path.splitext(args.dataFile[0])
#     (dataDir,) = os.path.split(args.dataFile[0])
#     # Set the floating point precision
#     nBits = 64

#     # Read in the data. Don't parse until inside the first function.
#     data = np.loadtxt(args.dataFile[0], unpack=True, dtype="float" + str(nBits))

#     # Run (modified) RM-synthesis on the spectra
#     mDict, aDict = run_adjoint_rmsynth(
#         data=data,
#         polyOrd=args.polyOrd,
#         phiMax_radm2=args.phiMax_radm2,
#         dPhi_radm2=args.dPhi_radm2,
#         nSamples=args.nSamples,
#         weightType=args.weightType,
#         fitRMSF=args.fitRMSF,
#         noStokesI=args.noStokesI,
#         nBits=nBits,
#         showPlots=args.showPlots,
#         debug=args.debug,
#         verbose=args.verbose,
#         units=args.units,
#     )

#     if args.saveOutput:
#         saveOutput(mDict, aDict, prefixOut, args.verbose)


# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     main()



# #!/usr/bin/env python
# # =============================================================================#
# #                                                                             #
# # NAME:     do_RMsynth_3D.py                                                  #
# #                                                                             #
# # PURPOSE:  Run RM-synthesis on a Stokes Q & U cubes.                         #
# #                                                                             #
# # MODIFIED: 7-March-2019 by J. West                                           #
# # MODIFIED: 23-October-2019 by A. Thomson                                     #
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2016 Cormac R. Purcell                                        #
# #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#

# import gc
# import math as m
# import os
# import sys
# import time

# import astropy.io.fits as pf
# import numpy as np

# from rm_lite.utils.fitting import interp_images
# from rm_lite.utils.synthesis import do_rmsynth_planes, get_rmsf_planes

# if sys.version_info.major == 2:
#     print("RM-tools will no longer run with Python 2! Please use Python 3.")
#     exit()

# C = 2.997924538e8  # Speed of light [m/s]

# # -----------------------------------------------------------------------------#


# def run_rmsynth(
#     dataQ,
#     dataU,
#     freqArr_Hz,
#     dataI=None,
#     rmsArr=None,
#     phiMax_radm2=None,
#     dPhi_radm2=None,
#     nSamples=10.0,
#     weightType="uniform",
#     fitRMSF=False,
#     nBits=32,
#     verbose=True,
#     not_rmsynth=False,
#     not_rmsf=False,
#     log=print,
#     super_resolution=False,
# ):
#     """Run RM-synthesis on 2/3D data.

#     Args:
#       dataQ (array_like): Stokes Q intensity cube.
#       dataU (array_like): Stokes U intensity cube.
#       freqArr_Hz (array_like): Frequency of each channel in Hz.

#     Kwargs:
#         dataI (array_like): Model cube of Stokes I spectra (see do_fitIcube).
#         rmsArr (array_like): Cube of uncertainty spectra.
#         phiMax_radm2 (float): Maximum absolute Faraday depth (rad/m^2).
#         dPhi_radm2 (float): Faraday depth channel size (rad/m^2).
#         nSamples (float): Number of samples across the RMSF.
#            weightType (str): Can be "variance" or "uniform"
#             "variance" -- Weight by inverse variance.
#             "uniform" -- Weight uniformly (i.e. with 1s)
#         fitRMSF (bool): Fit a Gaussian to the RMSF?
#         nBits (int): Precision of floating point numbers.
#         verbose (bool): Verbosity.
#         not_rmsynth (bool): Just do RMSF and ignore RM synthesis?
#         not_rmsf (bool): Just do RM synthesis and ignore RMSF? -- one of these must be False
#         log (function): Which logging function to use.

#     Returns:
#       dataArr (list): FDF and RMSF information
#         if not_rmsf:
#             dataArr = [FDFcube, phiArr_radm2, lam0Sq_m2, lambdaSqArr_m2]

#         else:
#             dataArr = [FDFcube, phiArr_radm2, RMSFcube, phi2Arr_radm2, fwhmRMSFCube,fitStatArr, lam0Sq_m2, lambdaSqArr_m2]


#     """
#     if not_rmsynth and not_rmsf:
#         log(
#             "Err: both RM synthesis and RMSF computation not requested?\n"
#             + "Please make sure either not_rmsynth or not_rmsf is False"
#         )
#         sys.exit()

#     # Sanity check on header dimensions

#     if not str(dataQ.shape) == str(dataU.shape):
#         log(
#             "Err: unequal dimensions: Q = "
#             + str(dataQ.shape)
#             + ", U = "
#             + str(dataU.shape)
#             + "."
#         )
#         sys.exit()

#     # Check dimensions of Stokes I cube, if present
#     if dataI is not None:
#         if not str(dataI.shape) == str(dataQ.shape):
#             log(
#                 "Err: unequal dimensions: Q = "
#                 + str(dataQ.shape)
#                 + ", I = "
#                 + str(dataI.shape)
#                 + "."
#             )
#             sys.exit()

#     # Default data types
#     dtFloat = "float" + str(nBits)
#     dtComplex = "complex" + str(2 * nBits)

#     lambdaSqArr_m2 = np.power(C / freqArr_Hz, 2.0)

#     lambdaSqRange_m2 = np.nanmax(lambdaSqArr_m2) - np.nanmin(lambdaSqArr_m2)
#     dLambdaSqMin_m2 = np.nanmin(np.abs(np.diff(lambdaSqArr_m2)))
#     dLambdaSqMax_m2 = np.nanmax(np.abs(np.diff(lambdaSqArr_m2)))

#     # Set the Faraday depth range
#     if not super_resolution:
#         fwhmRMSF_radm2 = 3.8 / lambdaSqRange_m2  # Dickey+2019 theoretical RMSF width
#     else:  # If super resolution, use R&C23 theoretical width
#         fwhmRMSF_radm2 = 2.0 / (np.nanmax(lambdaSqArr_m2) + np.nanmin(lambdaSqArr_m2))
#     if dPhi_radm2 is None:
#         dPhi_radm2 = fwhmRMSF_radm2 / nSamples
#     if phiMax_radm2 is None:
#         phiMax_radm2 = m.sqrt(3.0) / dLambdaSqMax_m2
#         phiMax_radm2 = max(
#             phiMax_radm2, fwhmRMSF_radm2 * 10.0
#         )  # Force the minimum phiMax to 10 FWHM

#     # Faraday depth sampling. Zero always centred on middle channel
#     nChanRM = int(round(abs((phiMax_radm2 - 0.0) / dPhi_radm2))) * 2 + 1
#     startPhi_radm2 = -(nChanRM - 1.0) * dPhi_radm2 / 2.0
#     stopPhi_radm2 = +(nChanRM - 1.0) * dPhi_radm2 / 2.0
#     phiArr_radm2 = np.linspace(startPhi_radm2, stopPhi_radm2, int(nChanRM))
#     phiArr_radm2 = phiArr_radm2.astype(dtFloat)
#     if verbose:
#         log(
#             "PhiArr = %.2f to %.2f by %.2f (%d chans)."
#             % (phiArr_radm2[0], phiArr_radm2[-1], float(dPhi_radm2), nChanRM)
#         )

#     # Calculate the weighting as 1/sigma^2 or all 1s (uniform)
#     if weightType == "variance" and rmsArr is not None:
#         weightArr = 1.0 / np.power(rmsArr, 2.0)
#     else:
#         weightType = "uniform"
#         weightArr = np.ones(freqArr_Hz.shape, dtype=dtFloat)
#     if verbose:
#         log("Weight type is '%s'." % weightType)

#     startTime = time.time()

#     # Read the Stokes I model and divide into the Q & U data
#     if dataI is not None:
#         with np.errstate(divide="ignore", invalid="ignore"):
#             qArr = np.true_divide(dataQ, dataI)
#             uArr = np.true_divide(dataU, dataI)
#     else:
#         qArr = dataQ
#         uArr = dataU

#     # Perform RM-synthesis on the cube
#     if not not_rmsynth:
#         FDFcube, lam0Sq_m2 = do_rmsynth_planes(
#             dataQ=qArr,
#             dataU=uArr,
#             lambdaSqArr_m2=lambdaSqArr_m2,
#             phiArr_radm2=phiArr_radm2,
#             weightArr=weightArr,
#             nBits=32,
#             lam0Sq_m2=0 if super_resolution else None,
#         )
#     else:
#         # need lambda0 for RMSF calculation
#         lam0Sq_m2 = 0 if super_resolution else None

#     # Calculate the Rotation Measure Spread Function cube
#     if not not_rmsf:
#         RMSFcube, phi2Arr_radm2, fwhmRMSFCube, fitStatArr, lam0Sq_m2 = get_rmsf_planes(
#             lambdaSqArr_m2=lambdaSqArr_m2,
#             phiArr_radm2=phiArr_radm2,
#             weightArr=weightArr,
#             mskArr=~np.isfinite(dataQ),
#             lam0Sq_m2=lam0Sq_m2,
#             double=True,
#             fitRMSF=fitRMSF or super_resolution,
#             fitRMSFreal=super_resolution,
#             nBits=32,
#             verbose=verbose,
#             log=log,
#         )
#     endTime = time.time()
#     cputime = endTime - startTime
#     if verbose:
#         log("> RM-synthesis completed in %.2f seconds." % cputime)

#     # Determine the Stokes I value at lam0Sq_m2 from the Stokes I model
#     # Note: the Stokes I model MUST be continuous throughout the cube,
#     # i.e., no NaNs as the amplitude at freq0_Hz is interpolated from the
#     # nearest two planes.
#     with np.errstate(divide="ignore", invalid="ignore"):
#         freq0_Hz = (
#             np.true_divide(C, m.sqrt(lam0Sq_m2))
#             if lam0Sq_m2 > 0
#             else np.nanmean(freqArr_Hz)
#         )

#     if dataI is not None:
#         idx = np.abs(freqArr_Hz - freq0_Hz).argmin()
#         if freqArr_Hz[idx] < freq0_Hz:
#             Ifreq0Arr = interp_images(dataI[idx, :, :], dataI[idx + 1, :, :], f=0.5)
#         elif freqArr_Hz[idx] > freq0_Hz:
#             Ifreq0Arr = interp_images(dataI[idx - 1, :, :], dataI[idx, :, :], f=0.5)
#         else:
#             Ifreq0Arr = dataI[idx, :, :]

#         # Multiply the dirty FDF by Ifreq0 to recover the PI
#         FDFcube *= Ifreq0Arr

#     if not_rmsf:  # only RMsynth
#         dataArr = [FDFcube, phiArr_radm2, lam0Sq_m2, lambdaSqArr_m2]
#     elif not_rmsynth:  # only RMSF
#         dataArr = [RMSFcube, phi2Arr_radm2, fwhmRMSFCube, fitStatArr, lam0Sq_m2]
#     else:  # both have been computed
#         dataArr = [
#             FDFcube,
#             phiArr_radm2,
#             RMSFcube,
#             phi2Arr_radm2,
#             fwhmRMSFCube,
#             fitStatArr,
#             lam0Sq_m2,
#             lambdaSqArr_m2,
#         ]

#     return dataArr


# def writefits(
#     dataArr,
#     headtemplate,
#     fitRMSF=False,
#     prefixOut="",
#     outDir="",
#     nBits=32,
#     write_separate_FDF=True,
#     not_rmsynth=False,
#     not_rmsf=False,
#     do_peakmaps=True,
#     verbose=False,
#     log=print,
# ):
#     """Write data to disk in FITS

#     Args:
#       dataArr (list): FDF and/or RMSF information
#         if not_rmsf: # only RMsynth
#             dataArr = [FDFcube, phiArr_radm2, lam0Sq_m2, lambdaSqArr_m2]
#         elif not_rmsynth: # only RMSF
#             dataArr = [RMSFcube, phi2Arr_radm2, fwhmRMSFCube, fitStatArr, lam0Sq_m2]
#         else: # both
#             dataArr = [FDFcube, phiArr_radm2, RMSFcube, phi2Arr_radm2, fwhmRMSFCube,fitStatArr, lam0Sq_m2, lambdaSqArr_m2]

#         headtemplate: FITS header template

#     Kwargs:
#         fitRMSF (bool): Fit a Gaussian to the RMSF?
#         prefixOut (str): Prefix for filenames.
#         outDir (str): Directory to save files.
#         write_separate_FDF (bool): Write Q, U, and PI separately?
#         verbose (bool): Verbosity.
#         not_rmsynth (bool): Just do RMSF and ignore RM synthesis?
#         not_rmsf (bool): Just do RM synthesis and ignore RMSF? -- one of these must be False
#         do_peakmaps (bool): Compute and write peak RM and peak intensity?
#         log (function): Which logging function to use.


#     Output files:
#         Default:
#         FDF_maxPI.fits: 2D map of peak polarized intensity per pixel.
#         FDF_peakRM.fits: 2D map of Faraday depth of highest peak, per pixel.

#         write_separate_FDF=True: [default]
#         FDF_dirty.fits is split into three constituent components:
#             FDF_real_dirty.fits: Stokes Q
#             FDF_im_dirty.fits: Stokes U
#             FDF_tot_dirty.fits: Polarizd Intensity (sqrt(Q^2+U^2))
#             RMSF_real.fits: Real/Stokes Q component of RMSF
#             RMSF_im.fits: Imaginary/Stokes U component of RMSF
#             RM_tot.fits: polarized intensity view of RMSF
#             RMSF_FWHM.fits: 2D map of width of RMSF main lobe

#         write_separate_FDF=False:
#             FDF_dirty.fits: FDF, in 3 extensions: Q,U, and PI.
#             RMSF.fits: 4 extensions; first 3 are RMSF cubes [Q, U, PI]
#                                  4th is 2D map of RMSF FWHM.


#     """
#     if not_rmsynth and not_rmsf:
#         log(
#             "Err: both RM synthesis and RMSF computation not done?\n"
#             + "Please make sure either not_rmsynth or not_rmsf is False"
#         )
#         sys.exit()

#     if not_rmsf:
#         FDFcube, phiArr_radm2, lam0Sq_m2, lambdaSqArr_m2 = dataArr
#         if verbose:
#             log("Saving the dirty FDF and ancillary FITS files.")
#     elif not_rmsynth:
#         RMSFcube, phi2Arr_radm2, fwhmRMSFCube, fitStatArr, lam0Sq_m2 = dataArr
#         if verbose:
#             log("Saving the RMSF and ancillary FITS files.")
#     else:
#         (
#             FDFcube,
#             phiArr_radm2,
#             RMSFcube,
#             phi2Arr_radm2,
#             fwhmRMSFCube,
#             fitStatArr,
#             lam0Sq_m2,
#             lambdaSqArr_m2,
#         ) = dataArr
#         if verbose:
#             log("Saving the dirty FDF, RMSF and ancillary FITS files.")

#     # Default data types
#     dtFloat = "float" + str(nBits)
#     dtComplex = "complex" + str(2 * nBits)

#     # Make a copy of the Q header and alter frequency-axis as Faraday depth
#     header = headtemplate.copy()
#     Ndim = header["NAXIS"]
#     freq_axis = Ndim  # If frequency axis not found, assume it's the last one.
#     # Check for frequency axes. Because I don't know what different formatting
#     # I might get ('FREQ' vs 'OBSFREQ' vs 'Freq' vs 'Frequency'), convert to
#     # all caps and check for 'FREQ' anywhere in the axis name.
#     for i in range(1, Ndim + 1):
#         try:
#             if "FREQ" in header["CTYPE" + str(i)].upper():
#                 freq_axis = i
#         except:
#             pass  # The try statement is needed for if the FITS header does not
#             # have CTYPE keywords.

#     header["CTYPE" + str(freq_axis)] = ("FDEP", "Faraday depth (linear)")
#     header["CUNIT" + str(freq_axis)] = "rad/m^2"
#     if not np.isfinite(lam0Sq_m2):
#         lam0Sq_m2 = 0.0
#     header["LAMSQ0"] = (lam0Sq_m2, "Lambda^2_0, in m^2")
#     if "DATAMAX" in header:
#         del header["DATAMAX"]
#     if "DATAMIN" in header:
#         del header["DATAMIN"]

#     if outDir == "":  # To prevent code breaking if file is in current directory
#         outDir = "."

#     # Re-add any initially removed degenerate axes (to match with FITS header)
#     # NOTE THIS HAS NOT BEEN RIGOROUSLY TESTED!!!
#     output_axes = []
#     for i in range(1, Ndim + 1):
#         output_axes.append(header["NAXIS" + str(i)])  # Get FITS dimensions
#     del output_axes[
#         freq_axis - 1
#     ]  # Remove frequency axis (since it's first in the array)
#     output_axes.reverse()  # To get into numpy order.

#     if "BUNIT" in header:
#         header["BUNIT"] = header["BUNIT"] + "/RMSF"

#     # Save the FDF
#     if not not_rmsynth:
#         header["NAXIS" + str(freq_axis)] = phiArr_radm2.size
#         header["CDELT" + str(freq_axis)] = (
#             np.diff(phiArr_radm2)[0],
#             "[rad/m^2] Coordinate increment at reference point",
#         )
#         header["CRPIX" + str(freq_axis)] = phiArr_radm2.size // 2 + 1
#         header["CRVAL" + str(freq_axis)] = (
#             phiArr_radm2[phiArr_radm2.size // 2],
#             "[rad/m^2] Coordinate value at reference point",
#         )

#         # Put frequency axis first, and reshape to add degenerate axes:
#         FDFcube = np.reshape(FDFcube, [FDFcube.shape[0]] + output_axes)
#         # Move Faraday depth axis to appropriate position to match header.
#         FDFcube = np.moveaxis(FDFcube, 0, Ndim - freq_axis)

#         if write_separate_FDF:  # more memory efficient as well
#             header = _setStokes(header, "Q")
#             hdu0 = pf.PrimaryHDU(FDFcube.real.astype(dtFloat), header)
#             fitsFileOut = outDir + "/" + prefixOut + "FDF_real_dirty.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu0.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             del hdu0
#             gc.collect()

#             header = _setStokes(header, "U")
#             hdu1 = pf.PrimaryHDU(FDFcube.imag.astype(dtFloat), header)
#             fitsFileOut = outDir + "/" + prefixOut + "FDF_im_dirty.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu1.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             del hdu1
#             gc.collect()

#             header = _setStokes(
#                 header, "PI"
#             )  # Sets Stokes axis to zero, which is a non-standard value.
#             del header["STOKES"]
#             hdu2 = pf.PrimaryHDU(np.abs(FDFcube).astype(dtFloat), header)
#             fitsFileOut = outDir + "/" + prefixOut + "FDF_tot_dirty.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu2.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             del hdu2
#             gc.collect()

#         else:
#             header = _setStokes(header, "Q")
#             hdu0 = pf.PrimaryHDU(FDFcube.real.astype(dtFloat), header)
#             header = _setStokes(header, "U")
#             hdu1 = pf.ImageHDU(FDFcube.imag.astype(dtFloat), header)
#             header = _setStokes(
#                 header, "PI"
#             )  # Sets Stokes axis to zero, which is a non-standard value.
#             del header["STOKES"]
#             hdu2 = pf.ImageHDU(np.abs(FDFcube).astype(dtFloat), header)

#             # Save the dirty FDF
#             fitsFileOut = outDir + "/" + prefixOut + "FDF_dirty.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hduLst = pf.HDUList([hdu0, hdu1, hdu2])
#             hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             hduLst.close()

#     # Save the RMSF
#     if not not_rmsf:
#         # Put frequency axis first, and reshape to add degenerate axes:
#         RMSFcube = np.reshape(RMSFcube, [RMSFcube.shape[0]] + output_axes)
#         # Move Faraday depth axis to appropriate position to match header.
#         RMSFcube = np.moveaxis(RMSFcube, 0, Ndim - freq_axis)

#         # Header for outputs that are RMSF
#         header["NAXIS" + str(freq_axis)] = phi2Arr_radm2.size
#         header["CDELT" + str(freq_axis)] = (
#             np.diff(phi2Arr_radm2)[0],
#             "[rad/m^2] Coordinate increment at reference point",
#         )
#         header["CRPIX" + str(freq_axis)] = phi2Arr_radm2.size // 2 + 1
#         header["CRVAL" + str(freq_axis)] = (
#             phi2Arr_radm2[phi2Arr_radm2.size // 2],
#             "[rad/m^2] Coordinate value at reference point",
#         )
#         header["BUNIT"] = ""

#         rmheader = header.copy()
#         rmheader["BUNIT"] = "rad/m^2"
#         if "BTYPE" in rmheader:
#             del rmheader["BTYPE"]
#         # Because there can be problems with different axes having different FITS keywords,
#         # don't try to remove the FD axis, but just make it degenerate.
#         # Also requires np.expand_dims to set the correct NAXIS.
#         rmheader["NAXIS" + str(freq_axis)] = 1
#         rmheader["CTYPE" + str(freq_axis)] = (
#             "DEGENERATE",
#             "Axis left in to avoid FITS errors",
#         )
#         rmheader["CUNIT" + str(freq_axis)] = ""
#         rmheader["CRVAL" + str(freq_axis)] = 0  # doesn't mean anything
#         stokes_axis = None
#         for axis in range(1, rmheader["NAXIS"] + 1):
#             if "STOKES" in rmheader[f"CTYPE{axis}"]:
#                 stokes_axis = axis
#         if stokes_axis is not None:
#             rmheader[f"CTYPE{stokes_axis}"] = (
#                 "DEGENERATE",
#                 "Axis left in to avoid FITS errors",
#             )

#         if write_separate_FDF:  # more memory efficient as well
#             header = _setStokes(header, "Q")
#             hdu0 = pf.PrimaryHDU(RMSFcube.real.astype(dtFloat), header)
#             fitsFileOut = outDir + "/" + prefixOut + "RMSF_real.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu0.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             del hdu0
#             gc.collect()

#             header = _setStokes(header, "U")
#             hdu1 = pf.PrimaryHDU(RMSFcube.imag.astype(dtFloat), header)
#             fitsFileOut = outDir + "/" + prefixOut + "RMSF_im.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu1.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             del hdu1
#             gc.collect()

#             header = _setStokes(
#                 header, "PI"
#             )  # Sets Stokes axis to zero, which is a non-standard value.
#             del header["STOKES"]
#             hdu2 = pf.PrimaryHDU(np.abs(RMSFcube).astype(dtFloat), header)
#             fitsFileOut = outDir + "/" + prefixOut + "RMSF_tot.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu2.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             del hdu2
#             gc.collect()

#             hdu3 = pf.PrimaryHDU(
#                 np.expand_dims(fwhmRMSFCube.astype(dtFloat), axis=0), rmheader
#             )
#             fitsFileOut = outDir + "/" + prefixOut + "RMSF_FWHM.fits"
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hdu3.writeto(fitsFileOut, output_verify="fix", overwrite=True)

#         else:
#             header = _setStokes(header, "Q")
#             hdu0 = pf.PrimaryHDU(RMSFcube.real.astype(dtFloat), header)
#             header = _setStokes(header, "U")
#             hdu1 = pf.ImageHDU(RMSFcube.imag.astype(dtFloat), header)
#             header = _setStokes(
#                 header, "PI"
#             )  # Sets Stokes axis to zero, which is a non-standard value.
#             del header["STOKES"]
#             hdu2 = pf.ImageHDU(np.abs(RMSFcube).astype(dtFloat), header)
#             hdu3 = pf.ImageHDU(
#                 np.expand_dims(fwhmRMSFCube.astype(dtFloat), axis=0), rmheader
#             )

#             fitsFileOut = outDir + "/" + prefixOut + "RMSF.fits"
#             hduLst = pf.HDUList([hdu0, hdu1, hdu2, hdu3])
#             if verbose:
#                 log("> %s" % fitsFileOut)
#             hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#             hduLst.close()

#     if not not_rmsynth and do_peakmaps:
#         ## Note that peaks are computed from the sampled functions
#         ## might be better to fit the FDF and compute the peak.
#         ## See RMpeakfit_3D.py

#         # Because there can be problems with different axes having different FITS keywords,
#         # don't try to remove the FD axis, but just make it degenerate.
#         # Also requires np.expand_dims to set the correct NAXIS.
#         # Generate peak maps:

#         maxPI, peakRM = create_peak_maps(FDFcube, phiArr_radm2, Ndim - freq_axis)
#         # Save a maximum polarised intensity map
#         if "BUNIT" in headtemplate:
#             header["BUNIT"] = headtemplate["BUNIT"]
#         header["NAXIS" + str(freq_axis)] = 1
#         header["CTYPE" + str(freq_axis)] = (
#             "DEGENERATE",
#             "Axis left in to avoid FITS errors",
#         )
#         header["CUNIT" + str(freq_axis)] = ""

#         # Header for output that are RM maps (peakRM, RMSF_FWHM, maxPI)
#         header["NAXIS" + str(freq_axis)] = 1
#         header["CRVAL" + str(freq_axis)] = (
#             phiArr_radm2[0],
#             "[rad/m^2] Coordinate value at reference point",
#         )
#         if "DATAMAX" in header:
#             del header["DATAMAX"]
#         if "DATAMIN" in header:
#             del header["DATAMIN"]

#         # Generate peak maps:

#         maxPI, peakRM = create_peak_maps(FDFcube, phiArr_radm2, Ndim - freq_axis)
#         # Save a maximum polarised intensity map
#         if "BUNIT" in headtemplate:
#             header["BUNIT"] = headtemplate["BUNIT"]
#         header["NAXIS" + str(freq_axis)] = 1
#         header["CTYPE" + str(freq_axis)] = (
#             "DEGENERATE",
#             "Axis left in to avoid FITS errors",
#         )
#         header["CUNIT" + str(freq_axis)] = ""

#         stokes_axis = None
#         for axis in range(1, header["NAXIS"] + 1):
#             if "STOKES" in header[f"CTYPE{axis}"]:
#                 stokes_axis = axis
#         if stokes_axis is not None:
#             header[f"CTYPE{stokes_axis}"] = (
#                 "DEGENERATE",
#                 "Axis left in to avoid FITS errors",
#             )

#         fitsFileOut = outDir + "/" + prefixOut + "FDF_maxPI.fits"
#         if verbose:
#             log("> %s" % fitsFileOut)
#         pf.writeto(
#             fitsFileOut,
#             np.expand_dims(maxPI.astype(dtFloat), axis=0),
#             header,
#             overwrite=True,
#             output_verify="fix",
#         )
#         # Save a peak RM map
#         fitsFileOut = outDir + "/" + prefixOut + "FDF_peakRM.fits"
#         header["BUNIT"] = "rad/m^2"
#         header["BTYPE"] = "FDEP"
#         if verbose:
#             log("> %s" % fitsFileOut)
#         pf.writeto(
#             fitsFileOut,
#             np.expand_dims(peakRM, axis=0),
#             header,
#             overwrite=True,
#             output_verify="fix",
#         )

#     #   #Cameron: I've removed the moment 1 map for now because I don't think it's properly/robustly defined.
#     #    # Save an RM moment-1 map
#     #    fitsFileOut = outDir + "/" + prefixOut + "FDF_mom1.fits"
#     #    header["BUNIT"] = "rad/m^2"
#     #    mom1FDFmap = (np.nansum(np.moveaxis(np.abs(FDFcube),FDFcube.ndim-freq_axis,FDFcube.ndim-1) * phiArr_radm2, FDFcube.ndim-1)
#     #                  /np.nansum(np.abs(FDFcube), FDFcube.ndim-freq_axis))
#     #    mom1FDFmap = mom1FDFmap.astype(dtFloat)
#     #    if(verbose): log("> %s" % fitsFileOut)
#     #    pf.writeto(fitsFileOut, mom1FDFmap, header, overwrite=True,
#     #               output_verify="fix")


# def _setStokes(header, stokes):
#     """Check if header has Stokes axis. If so, set to correct numerical value
#     (if IQUV). If not a valid Stokes parameter, sets to zero.
#     Adds Stokes keyword regardless. Returns new, updated header.
#     """
#     stokes_dict = {"I": 1, "Q": 2, "U": 3, "V": 4}

#     outheader = header.copy()
#     stokes_axis = None
#     for axis in range(1, header["NAXIS"] + 1):
#         if "STOKES" in header[f"CTYPE{axis}"]:
#             stokes_axis = axis

#     if stokes_axis is not None:
#         outheader[f"CRPIX{stokes_axis}"] = 1.0
#         outheader[f"CRVAL{stokes_axis}"] = stokes_dict.get(stokes.upper(), 0)
#     outheader["STOKES"] = stokes

#     return outheader


# def create_peak_maps(FDFcube, phiArr_radm2, phi_axis=0):
#     """Finds the location and amplitude of the highest peak in the FDF (pixelwise)
#     and returns maps of those parameters. Does not fit the peak, only finds
#     the location in terms of the quantized Faraday depth slices of the cube.
#     Used to produce the maxPI and peakRM maps.
#     Inputs:
#         FDFcube: output cube from run_rmsynth
#         phiArr_radm2: array of Faraday depth values, from run_rmsynth
#         phi_axis (int): number of the axis for Faraday depth (in python order,
#                          not FITS order). Defaults to zero (first axis).
#     Returns:
#         maxPI: array of same dimensions as FDFcube exceppt collapsed along
#                 first (Faraday depth) axis, containing the maximum polarized
#                 intensity for each pixel
#         peakRM: as maxPI, but with the Faraday depth location of the peak
#     """

#     maxPI = np.max(np.abs(FDFcube), axis=phi_axis)
#     peakRM_indices = np.argmax(np.abs(FDFcube), axis=phi_axis)
#     peakRM = phiArr_radm2[peakRM_indices]
#     # Check for pixels with all NaNs across FD
#     # Write peakRM as NaN (otherwise it takes the first entry in phiArr)
#     nan_mask = np.all(np.isnan(FDFcube), axis=phi_axis)
#     peakRM[nan_mask] = np.nan

#     return maxPI, peakRM


# def find_freq_axis(header):
#     """Finds the frequency axis in a FITS header.
#     Input: header: a Pyfits header object.
#     Returns the axis number (as recorded in the FITS file, **NOT** in numpy ordering.)
#     Returns 0 if the frequency axis cannot be found.
#     """
#     freq_axis = 0  # Default for 'frequency axis not identified'
#     # Check for frequency axes. Because I don't know what different formatting
#     # I might get ('FREQ' vs 'OBSFREQ' vs 'Freq' vs 'Frequency'), convert to
#     # all caps and check for 'FREQ' anywhere in the axis name.
#     for i in range(1, header["NAXIS"] + 1):  # Check each axis in turn.
#         try:
#             if "FREQ" in header["CTYPE" + str(i)].upper():
#                 freq_axis = i
#         except:
#             pass  # The try statement is needed for if the FITS header does not
#             # have CTYPE keywords.
#     return freq_axis


# def readFitsCube(file, verbose, log=print):
#     """The old version of this function could only accept 3 or 4 axis input
#     (and implicitly assumed that in the 4 axis case that axis 3 was degenerate).
#     I'm trying to somewhat generalize this, so that it will accept NAXIS=1..3
#     cases and automatically try to identify which axis is the frequency axis,
#     and will try to remove the degenerate axis in the 4D case.
#     Where it can't find the correct frequency axis, it will assume it is the
#     last one. It assumes any fourth or higher dimensions are degenerate (length 1)
#     and will remove them. If the higher dimensions are NOT degenerate (e.g., a
#     cube with all 4 Stokes), the code will fail (support may be added later?).
#     -Cameron (3 April 2019)
#     """
#     if not os.path.exists(file):
#         log("Err: File not found")

#     if verbose:
#         log("Reading " + file + " ...")
#     data = pf.getdata(file)
#     head = pf.getheader(file)
#     if verbose:
#         log("done.")

#     N_dim = head["NAXIS"]  # Get number of axes
#     if verbose:
#         print("Dimensions of the input cube are: ", end=" ")
#         for i in range(1, N_dim + 1):
#             print("NAXIS{} = {}".format(i, head["NAXIS" + str(i)]), end="  ")
#         print()

#     freq_axis = find_freq_axis(head)
#     # If the frequency axis isn't the last one, rotate the array until it is.
#     # Recall that pyfits reverses the axis ordering, so we want frequency on
#     # axis 0 of the numpy array.
#     if freq_axis != 0 and freq_axis != N_dim:
#         data = np.moveaxis(data, N_dim - freq_axis, 0)

#     if N_dim >= 4:
#         data = np.squeeze(data)  # Remove degenerate axes

#     if verbose:
#         print("Dimensions of the input array are: ", data.shape)

#     if data.ndim > 3:
#         raise Exception("Data cube has too many (non-degenerate) axes!")

#     return head, data


# def readFreqFile(file, verbose, log=print):
#     # Read the frequency vector and wavelength sampling
#     freqArr_Hz = np.loadtxt(file, dtype=float)
#     return freqArr_Hz


# # -----------------------------------------------------------------------------#
# def main():
#     import argparse

#     """
#     Start the function to perform RM-synthesis if called from the command line.
#     """

#     # Help string to be shown using the -h option
#     descStr = """
#     Run RM-synthesis on a pair of Stokes Q and U cubes (3D). This script
#     correctly deals with isolated clumps of flagged voxels in the cubes (NaNs).
#     Saves cubes containing the complex Faraday dispersion function (FDF), a
#     cube of double-size Rotation Measure Spread Functions, a peak Faraday
#     depth map, a first-moment map and a maximum polarised intensity map.
#     """

#     epilog_text = """
#     Output files:
#         Default:
#         FDF_real_dirty.fits: real (Stokes Q) component of the FDF
#         FDF_im_dirty.fits: imaginary (Stokes U) component of the FDF
#         FDF_tot_dirty.fits: polarized intnsity (Stokes P) component of the FDF
#         FDF_maxPI.fits: 2D map of peak polarized intensity per pixel.
#         FDF_peakRM.fits: 2D map of Faraday depth of highest peak, per pixel.
#         RMSF_real_dirty.fits: real (Stokes Q) component of the RMSF
#         RMSF_im_dirty.fits: imaginary (Stokes U) component of the RMSF
#         RMSF_tot_dirty.fits: polarized intnsity (Stokes P) component of the RMSF
#         RMSF_FWHM: 2D map of RMSF FWHM per pixel.

#         With -f flag, the 3 FDF cubes are combined in a single file with 3 extensions.
#             and the RMSF files are combined in a single file with 4 extensions.

#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr,
#         epilog=epilog_text,
#         formatter_class=argparse.RawTextHelpFormatter,
#     )
#     parser.add_argument(
#         "fitsQ",
#         metavar="StokesQ.fits",
#         nargs=1,
#         help="FITS cube containing Stokes Q data.",
#     )
#     parser.add_argument(
#         "fitsU",
#         metavar="StokesU.fits",
#         nargs=1,
#         help="FITS cube containing Stokes U data.",
#     )
#     parser.add_argument(
#         "freqFile",
#         metavar="freqs_Hz.dat",
#         nargs=1,
#         help="ASCII file containing the frequency vector.",
#     )
#     parser.add_argument(
#         "-i",
#         dest="fitsI",
#         default=None,
#         help="FITS cube containing Stokes I model [None].",
#     )
#     parser.add_argument(
#         "-n",
#         dest="noiseFile",
#         default=None,
#         help="Text file containing channel noise values [None].",
#     )
#     parser.add_argument(
#         "-w",
#         dest="weightType",
#         default="uniform",
#         help="weighting ['uniform'] (all 1s) or 'variance'.",
#     )
#     parser.add_argument(
#         "-t",
#         dest="fitRMSF",
#         action="store_true",
#         help="Fit a Gaussian to the RMSF [False]",
#     )
#     parser.add_argument(
#         "-l",
#         dest="phiMax_radm2",
#         type=float,
#         default=None,
#         help="Absolute max Faraday depth sampled (overrides NSAMPLES) [Auto].",
#     )
#     parser.add_argument(
#         "-d",
#         dest="dPhi_radm2",
#         type=float,
#         default=None,
#         help="Width of Faraday depth channel [Auto].",
#     )
#     parser.add_argument(
#         "-o",
#         dest="prefixOut",
#         default="",
#         help="Prefix to prepend to output files [None].",
#     )
#     parser.add_argument(
#         "-s",
#         dest="nSamples",
#         type=float,
#         default=5,
#         help="Number of samples across the FWHM RMSF.",
#     )
#     parser.add_argument(
#         "-f",
#         dest="write_separate_FDF",
#         action="store_false",
#         help="Store different Stokes as FITS extensions [False, store as separate files].",
#     )
#     parser.add_argument(
#         "-v", dest="verbose", action="store_true", help="Verbose [False]."
#     )
#     parser.add_argument(
#         "-R",
#         dest="not_RMSF",
#         action="store_true",
#         help="Skip calculation of RMSF? [False]",
#     )
#     parser.add_argument(
#         "-r",
#         "--super-resolution",
#         action="store_true",
#         help="Optimise the resolution of the RMSF (as per Rudnick & Cotton). ",
#     )
#     args = parser.parse_args()

#     # Sanity checks
#     for f in args.fitsQ + args.fitsU:
#         if not os.path.exists(f):
#             print("File does not exist: '%s'." % f)
#             sys.exit()
#     (dataDir,) = os.path.split(args.fitsQ[0])
#     verbose = args.verbose
#     if args.fitsI is not None:
#         dataI = readFitsCube(args.fitsI, verbose)[1]
#     else:
#         dataI = None
#     if args.noiseFile is not None:
#         rmsArr = readFreqFile(args.noiseFile, verbose)
#     else:
#         rmsArr = None

#     header, dataQ = readFitsCube(args.fitsQ[0], verbose)

#     # Run RM-synthesis on the cubes
#     dataArr = run_rmsynth(
#         dataQ=dataQ,
#         dataU=readFitsCube(args.fitsU[0], verbose)[1],
#         freqArr_Hz=readFreqFile(args.freqFile[0], verbose),
#         dataI=dataI,
#         rmsArr=rmsArr,
#         phiMax_radm2=args.phiMax_radm2,
#         dPhi_radm2=args.dPhi_radm2,
#         nSamples=args.nSamples,
#         weightType=args.weightType,
#         fitRMSF=args.fitRMSF,
#         nBits=32,
#         verbose=verbose,
#         not_rmsf=args.not_RMSF,
#         super_resolution=args.super_resolution,
#     )

#     # Write to files
#     writefits(
#         dataArr,
#         headtemplate=header,
#         fitRMSF=False,
#         prefixOut=args.prefixOut,
#         outDir=dataDir,
#         write_separate_FDF=args.write_separate_FDF,
#         not_rmsf=args.not_RMSF,
#         nBits=32,
#         verbose=verbose,
#     )


# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     main()



# #!/usr/bin/env python
# # =============================================================================#
# #                                                                             #
# # NAME:     do_fitIcube.py                                                    #
# #                                                                             #
# # PURPOSE:  Make a model Stokes I cube and a noise vector.                    #
# #                                                                             #
# # MODIFIED: 26-Feb-2017 by C. Purcell
# # MODIFIED: 18 January 2023 by Lerato Baidoo  (re-structured and optimized)   #
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2016 Cormac R. Purcell                                        #
# #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#

# import argparse
# import os
# import sys
# import time
# from functools import partial

# import astropy.io.fits as pf
# import numpy as np
# from tools_3d.make_freq_file import get_freq_arr
# from tqdm.contrib.concurrent import process_map
# from utils.util_FITS import strip_fits_dims

# from rm_lite.tools_3d.rmsynth import readFitsCube
# from rm_lite.utils.fitting import MAD, calculate_StokesI_model, fit_StokesI_model

# # -----------------------------------------------------------------------------#


# def main():
#     """
#     Start the make_model_I function if called from the command line.
#     """

#     # Help string to be shown using the -h option
#     descStr = """
#     Create a model Stokes I dataset by fitting a polynomial to emitting regions
#     above a cutoff threshold in the Stokes I cube. Also outputs a noise spectrum
#     with the Stokes I noise per channel.

#     NOTE: Each pixel is fit independently, so there are no protections in place
#     to ensure smoothness across the image-plane. Noise levels are estimated
#     per-channel using 2 passes of the MAD.
#     The source mask, if applied, is calculated per-pixel by looking for values
#     above a threshold (either an absolute intensity threshold or a multiple of
#     the noise) in any channel.
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr, formatter_class=argparse.RawTextHelpFormatter
#     )
#     parser.add_argument(
#         "fitsI",
#         metavar="StokesI.fits",
#         nargs=1,
#         help="FITS cube containing Stokes I data.",
#     )
#     parser.add_argument(
#         "-freqFile",
#         dest="freq_file",
#         default="",
#         type=str,
#         help="Path + ASCII file containing the frequency vector. If not provided,\nfrequencies are derived from fits header.",
#     )
#     parser.add_argument(
#         "-f",
#         dest="fit_function",
#         type=str,
#         default="log",
#         help="Stokes I fitting function: 'linear' or ['log'] polynomials.",
#     )
#     parser.add_argument(
#         "-p",
#         dest="polyOrd",
#         type=int,
#         default=2,
#         help="polynomial order to fit to I spectrum: 0-5 supported, 2 is default.\nSet to negative number to enable dynamic order selection.",
#     )
#     parser.add_argument(
#         "-n",
#         dest="num_cores",
#         type=int,
#         default=1,
#         help="Number of cores to use for multiprocessing. Default is 1.",
#     )
#     parser.add_argument(
#         "-m",
#         dest="apply_mask",
#         action="store_true",
#         help="Apply masking before spectral fitting. Default is False.",
#     )
#     parser.add_argument(
#         "-c",
#         dest="chunk_size",
#         type=int,
#         default=1,
#         help="Chunk size for multiprocessing. Default is 1.",
#     )
#     parser.add_argument(
#         "-t",
#         dest="threshold",
#         type=float,
#         default=-3,
#         help="Source masking threshold in flux units (if positive) or factors of per-channel sigma (if negative). Default is -3 (i.e. 3 sigma mask).",
#     )
#     parser.add_argument(
#         "-o",
#         dest="prefixOut",
#         default="",
#         help="Prefix to use for to output file names.",
#     )
#     parser.add_argument(
#         "-odir",
#         dest="outDir",
#         default="",
#         help="Output directory to save output files. If none, save inside input directory",
#     )
#     parser.add_argument(
#         "-v",
#         dest="verbose",
#         action="store_true",
#         help="turn on verbose messages [False].",
#     )
#     args = parser.parse_args()

#     # Sanity checks
#     for f in [args.fitsI[0]]:
#         if not os.path.exists(f):
#             print("File does not exist: '%s'." % f)
#             sys.exit()
#     out_directory = args.outDir
#     if not out_directory:
#         (out_directory,) = os.path.split(args.fitsI[0])

#     I_filename = args.fitsI[0]
#     datacube, headI = open_datacube(fitsI=I_filename, verbose=args.verbose)

#     # if frequency file is not provided, extract frequency information from the input fits header.
#     if args.freq_file:
#         freqArr_Hz = get_frequencies(
#             datacube=datacube, header=headI, freqFile=args.freq_file
#         )
#     else:
#         print("Frequency file not provided. Deriving frequencies from the fits header.")
#         freqArr_Hz = get_freq_arr(I_filename)

#     # Run polynomial fitting on the spectra
#     make_model_I(
#         datacube=datacube,
#         header=headI,
#         freqArr_Hz=freqArr_Hz,
#         polyOrd=args.polyOrd,
#         prefixOut=args.prefixOut,
#         outDir=out_directory,
#         nBits=32,
#         threshold=args.threshold,
#         apply_mask=args.apply_mask,
#         num_cores=args.num_cores,
#         chunk_size=args.chunk_size,
#         verbose=args.verbose,
#         fit_function=args.fit_function,
#     )


# # -----------------------------------------------------------------------------#


# def open_datacube(fitsI, verbose=False):
#     """Reads the image fits

#     Parameters:
#     fitsI : Input Stokes I cube fits image
#     verbose: If true, write logs

#     Returns:
#     datacube: Image data
#     header: Fits header
#     """

#     # Default data type

#     # Sanity check on header dimensions
#     print("Reading FITS cube header from '%s':" % fitsI)
#     header, datacube = readFitsCube(fitsI, verbose)

#     nDim = datacube.ndim
#     if nDim < 3 or nDim > 4:
#         print("Err: only 3 or 4 dimensions supported: D = %d." % header["NAXIS"])
#         sys.exit()

#     # freq_axis=find_freq_axis(header)
#     # #If the frequency axis isn't the last one, rotate the array until it is.
#     # #Recall that pyfits reverses the axis ordering, so we want frequency on
#     # #axis 0 of the numpy array.
#     # if freq_axis != 0 and freq_axis != nDim:
#     #     datacube=np.moveaxis(datacube,nDim-freq_axis,0)

#     return datacube, header


# def get_frequencies(datacube, header, freqFile):
#     """Reads a frequency file

#     Parameters:
#     datacube: Image cube data
#     header: A header of the input datacube
#     freqFile: A frequency file in text format

#     Returns:
#     freqArr_Hz: frequency array
#     """

#     nBits = np.abs(header["BITPIX"])
#     dtFloat = "float" + str(nBits)

#     nChan = datacube.shape[0]  # for now, assumes frequency is the first axis

#     # Read the frequency vector
#     print("Reading frequency vector from '%s'." % freqFile)
#     freqArr_Hz = np.loadtxt(freqFile, dtype=dtFloat)
#     if nChan != len(freqArr_Hz):
#         print("Err: frequency vector and frequency axis of cube unequal length.")
#         sys.exit()

#     return freqArr_Hz


# def cube_noise(datacube, header, freqArr_Hz, threshold=-5):
#     """Estimates noise of each channel in an image cube.

#     Parameters:
#     datacube: Input cube data
#     header : Header of a cube image
#     frequency: Frequency values of a cube image in Hz
#     threshold: Threshold to use for masking off pixels.
#             If the value is +ve, then it is taken as absolute value, and
#             -ve as sigma. E.g. threshold=-5, means consider pixels
#             with emission > noise_median - 5 * rms_noise.
#             Where the noise_median and rms_noise are obtained by computing
#             the median and MAD of the pixel emission > image median + 3 * image MAD.

#     Returns:
#     rms_Arr: An array containing rms values of each channel
#     mskSrc:  A 2D image data containing masking values (0s and 1s)
#     """
#     nBits = np.abs(header["BITPIX"])
#     dtFloat = "float" + str(nBits)
#     nChan = datacube.shape[0]

#     if nChan != len(freqArr_Hz):
#         print("Err: frequency vector and frequency axis of cube unequal length.")
#         sys.exit()

#     # Measure the RMS spectrum using 2 passes of MAD on each plane
#     # Determine which pixels have emission above the threshold
#     print("Measuring the per-channel noise and creating an emission mask")
#     rmsArr = np.zeros_like(freqArr_Hz)
#     medSky = np.zeros_like(freqArr_Hz)
#     mskSrc = np.zeros((header["NAXIS2"], header["NAXIS1"]), dtype=dtFloat)

#     # start = time.time()
#     for i in range(nChan):
#         dataPlane = datacube[i]
#         if threshold > 0:
#             idxSky = np.where(dataPlane < threshold)  # replaced cutoff with threshold
#         else:
#             idxSky = np.where(dataPlane)

#         # Pass 1
#         rmsTmp = MAD(dataPlane[idxSky])
#         medTmp = np.nanmedian(dataPlane[idxSky])

#         # Pass 2: use a fixed 3-sigma cutoff to mask off emission
#         idxSky = np.where(dataPlane < medTmp + rmsTmp * 3)
#         medSky[i] = np.nanmedian(dataPlane[idxSky])
#         rmsArr[i] = MAD(dataPlane[idxSky])

#         # When building final emission mask treat +ve threshold as absolute
#         # values and negative threshold as sigma values
#         if threshold > 0:
#             idxSrc = np.where(dataPlane > threshold)
#         else:
#             idxSrc = np.where(dataPlane > medSky[i] - 1 * rmsArr[i] * threshold)

#         mskSrc[idxSrc] += 1

#     # end = time.time()
#     # print(' For loop masking takes %.3fs'%(end-start))
#     return rmsArr, mskSrc


# def savefits_mask(data, header, outDir, prefixOut):
#     """Save the derived mask to a fits file

#     data:  2D data defining the mask.
#     header: header to describe the mask
#     outDir: directory to save the mask fits data
#     prefixOut: prefix to use on the output name
#     """

#     headMask = strip_fits_dims(header=header, minDim=2)
#     headMask["DATAMAX"] = 1
#     headMask["DATAMIN"] = 0
#     if "BUNIT" in headMask:
#         del headMask["BUNIT"]

#     mskArr = np.where(data > 0, 1.0, np.nan)
#     MaskfitsFile = os.path.join(outDir, prefixOut + "mask.fits")
#     print("> %s" % MaskfitsFile)
#     pf.writeto(MaskfitsFile, mskArr, headMask, output_verify="fix", overwrite=True)


# def savefits_Coeffs(data, dataerr, header, polyOrd, outDir, prefixOut):
#     """Save the derived coefficients to a fits file

#     data: 2D planes containing coeffs values.
#     dataerr: 2D planes containing error in coeffs values.
#     header: header to describe the coeffs and error in coeffs
#     polyOrd: the order of polynomial to fit
#     outDir: directory to save the (errors) coeffs fits data
#     prefixOut: prefix to use on the output name
#     """

#     header["BUNIT"] = ""
#     if "BTYPE" in header:
#         del header["BTYPE"]

#     for i in range(np.abs(polyOrd) + 1):
#         outname = os.path.join(outDir, prefixOut + "coeff" + str(i) + ".fits")
#         pf.writeto(outname, data[i], header, overwrite=True)

#         outname = os.path.join(outDir, prefixOut + "coeff" + str(i) + "err.fits")
#         pf.writeto(outname, dataerr[i], header, overwrite=True)


# def savefits_model_I(data, header, outDir, prefixOut):
#     """Save the derived Stokes cube model

#     data:  Stokes I cube model data
#     header: header to describe the model cube.
#     outDir: directory to save the model cube fits data
#     prefixOut: prefix to use on the output name
#     """

#     nDim = data.ndim
#     nBits = np.abs(header["BITPIX"])

#     headModelCube = strip_fits_dims(header=header, minDim=nDim)
#     headModelCube["NAXIS1"] = header["NAXIS1"]
#     headModelCube["NAXIS2"] = header["NAXIS2"]
#     headModelCube["NAXIS3"] = header["NAXIS3"]

#     nVoxels = header["NAXIS1"] * header["NAXIS2"] * header["NAXIS3"]
#     if nDim == 4:
#         headModelCube["NAXIS4"] = header["NAXIS4"]
#         nVoxels *= header["NAXIS4"]
#     while len(headModelCube) < (36 * 4 - 1):
#         headModelCube.append()

#     fitsModelFile = os.path.join(outDir, prefixOut + "model.i.fits")
#     headModelCube.tofile(fitsModelFile, overwrite=True)
#     with open(fitsModelFile, "rb+") as f:
#         f.seek(len(headModelCube.tostring()) + (nVoxels * int(nBits / 8)) - 1)
#         f.write(b"\0")
#     HDULst = pf.open(fitsModelFile, "update", memmap=True)
#     HDULst[0].data = data
#     HDULst.close()


# def fit_spectra_I(
#     Ispectrum, freqArr_Hz, rms_Arr, polyOrd, fit_function, nDetectPix, verbose=False
# ):
#     """Fits polynomial function to Stokes I data

#     xy: Position of pixel to fit (in pixels).
#         The xy consists of pixel number, x and y pixel position.
#     datacube:  Stokes I data cube to model.
#     freqArr_Hz: Frequency array in Hz.
#     rms_Arr: An array containing rms values of each channel.
#     polyOrd: the order of polynomial to fit.
#     fit_function: A type of function to fit.
#          It can be log or linear.
#     nDetectPix:  the total number of pixels to be fit.
#     """

#     pix_fit_result = fit_StokesI_model(
#         freqArr_Hz, Ispectrum, rms_Arr, polyOrd=polyOrd, fit_function=fit_function
#     )

#     pixImodel = calculate_StokesI_model(pix_fit_result, freqArr_Hz)

#     outs = dict()

#     outs["I"] = pixImodel.astype("float32")
#     outs["coeffs"] = pix_fit_result.params.astype("float32")
#     outs["coeffs_err"] = pix_fit_result.perror.astype("float32")
#     outs["chiSq"] = pix_fit_result.chiSq
#     outs["chiSqRed"] = pix_fit_result.chiSqRed
#     outs["nIter"] = pix_fit_result.nIter
#     outs["AIC"] = pix_fit_result.AIC
#     outs["covar"] = pix_fit_result.pcov
#     outs["reference_frequency_Hz"] = pix_fit_result.reference_frequency_Hz

#     return outs


# def make_model_I(
#     datacube,
#     header,
#     freqArr_Hz,
#     polyOrd=2,
#     nBits=32,
#     threshold=3,
#     num_cores=1,
#     chunk_size=1,
#     verbose=False,
#     fit_function="log",
#     apply_mask=False,
#     outDir=None,
#     prefixOut=None,
# ):
#     """Fits a polynomial function to Stokes I data, derives coefficients,
#        predicts model I, and save the respective fits file.

#     datacube:  Stokes I data cube
#     header: header of the data cube
#     freqArr_Hz: frequency values of the cube in Hz
#     polyOrd: the order of the polynomial to fit. 0-5 supported, 2 is default
#     fit_function: fit log or linear

#     num_cores: Number of cores to use for parallel processing
#     chunk_size: Chunk size for multiprocessing
#     verbose: Write to log
#     apply_mask: If true, a mask will be applied
#     threshold: Threshold to use for masking off pixels

#             If the value is +ve, then it is taken as absolute value, and
#             -ve as sigma. E.g. threshold=-5, means consider pixels
#             with emission > noise_median - 5 * rms_noise.
#             Where the noise_median and rms_noise are obtained by computing
#             the median and MAD of the pixel emission > image median + 3 * image MAD
#     outDir: Directory to save all outputs
#     prefixOut: Prefix name to use in all output names

#     Returns:
#     ModelIcube: Model I cube data array
#     Mask fits data
#     Coefficients fits data
#     Error in coefficients fits data
#     """
#     dtFloat = "float" + str(nBits)

#     rms_Arr, mskSrc = cube_noise(datacube, header, freqArr_Hz, threshold=threshold)

#     mskArr = np.where(mskSrc > 0, 1.0, np.nan)

#     if not apply_mask:
#         mskSrc = np.ones((header["naxis2"], header["naxis1"]), dtype=dtFloat)
#         mskArr = np.where(mskSrc > 0, 1.0, np.nan)

#     srcCoords = np.rot90(np.where(mskSrc > 0))

#     nPix = mskSrc.shape[-1] * mskSrc.shape[-2]
#     nDetectPix = len(srcCoords)

#     if verbose and apply_mask:
#         print(
#             "Emission present in %d spectra (%.1f percent)."
#             % (nDetectPix, (nDetectPix * 100.0 / nPix))
#         )

#     modelIcube = np.zeros_like(datacube)
#     modelIcube[:] = np.nan
#     results = []

#     coeffs = np.array([mskArr] * 6)
#     coeffs_error = np.array([mskArr] * 6)
#     reffreq = np.squeeze(np.array([mskArr]))

#     covars = np.array([[mskArr] * 6] * 6)
#     datacube = np.squeeze(datacube)

#     # Select only the spectra with emission
#     srcData = np.rot90(datacube[:, mskSrc > 0])

#     # Inform user job magnitude
#     startTime = time.time()

#     xy = list(zip(np.arange(0, len(srcCoords)), srcCoords[:, 0], srcCoords[:, 1]))

#     if verbose:
#         print("Fitting %d/%d spectra." % (nDetectPix, nPix))
#     if verbose:
#         print(
#             f"Using {num_cores} cores and chunksize {chunk_size} for parallel processing."
#         )

#     func = partial(
#         fit_spectra_I,
#         freqArr_Hz=freqArr_Hz,
#         rms_Arr=rms_Arr,
#         polyOrd=polyOrd,
#         fit_function=fit_function,
#         nDetectPix=nDetectPix,
#         verbose=verbose,
#     )
#     # Send each spectrum to a different core
#     results = process_map(
#         func,
#         srcData,
#         max_workers=num_cores,
#         chunksize=chunk_size,
#         disable=not verbose,
#         desc="Fitting spectra",
#         total=nDetectPix,
#     )

#     headcoeff = strip_fits_dims(header=header, minDim=2)
#     del headcoeff["BUNIT"]

#     endTime = time.time()
#     cputime = endTime - startTime
#     print("Fitting completed in %.2f seconds." % cputime)

#     for _, an in enumerate(xy):
#         i, x, y = an

#         modelIcube[:, x, y] = results[i]["I"]
#         reffreq[x, y] = results[i]["reference_frequency_Hz"]
#         covars[:, :, x, y] = results[i]["covar"]

#         for k, j, l in zip(
#             range(len(coeffs)), results[i]["coeffs"], results[i]["coeffs_err"]
#         ):
#             coeffs[5 - k, x, y] = j
#             coeffs_error[5 - k, x, y] = l

#     headcoeff["HISTORY"] = "Stokes I model fitted by RM-Tools"
#     if polyOrd < 0:
#         headcoeff["HISTORY"] = (
#             f"Fit model is dynamic order {fit_function}-polynomial, max order {-polyOrd}"
#         )
#     else:
#         headcoeff["HISTORY"] = f"Fit model is {polyOrd}-order {fit_function}-polynomial"

#     if verbose:
#         print("Saving mask image.")
#     savefits_mask(data=mskSrc, header=headcoeff, outDir=outDir, prefixOut=prefixOut)

#     if verbose:
#         print("Saving model I coefficients.")
#     savefits_Coeffs(
#         data=coeffs,
#         dataerr=coeffs_error,
#         header=headcoeff,
#         polyOrd=polyOrd,
#         outDir=outDir,
#         prefixOut=prefixOut,
#     )

#     # Save frequency map
#     head_freq = headcoeff.copy()
#     head_freq["BUNIT"] = "Hz"
#     if "BTYPE" in headcoeff:
#         del headcoeff["BTYPE"]

#     outname = os.path.join(outDir, prefixOut + "reffreq.fits")
#     pf.writeto(outname, reffreq, head_freq, overwrite=True)

#     # Save covariance maps -- these are necessary if/when converting the model
#     # reference frequency.
#     # Structure will be a single file as a 4D cube, with the 3rd and 4th dimensions
#     # iterating over the two axes of the covariance matrix.
#     head_covar = headcoeff.copy()
#     head_covar["NAXIS"] = 4
#     head_covar["NAXIS3"] = 6
#     head_covar["NAXIS4"] = 6
#     head_covar["CTYPE3"] = "INDEX"
#     head_covar["CTYPE4"] = "INDEX"
#     head_covar["CRVAL3"] = 0
#     head_covar["CRVAL4"] = 0
#     head_covar["CDELT3"] = 1
#     head_covar["CDELT4"] = 1
#     head_covar["CRPIX3"] = 1
#     head_covar["CRPIX4"] = 1
#     head_covar["CUNIT3"] = ""
#     head_covar["CUNIT4"] = ""

#     outname = os.path.join(outDir, prefixOut + "covariance.fits")
#     pf.writeto(outname, covars, head_covar, overwrite=True)

#     if verbose:
#         print("Saving model I cube image. ")
#     savefits_model_I(data=modelIcube, header=header, outDir=outDir, prefixOut=prefixOut)

#     np.savetxt(os.path.join(outDir, prefixOut + "noise.dat"), rms_Arr)

#     return modelIcube


# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     main()



"""Tests for the synthesis utilities"""

from __future__ import annotations

import numpy as np
from matplotlib.pylab import Generator
from rm_lite.utils.synthesis import (
    freq_to_lambda2,
    lambda2_to_freq,
    make_double_phi_arr,
    make_phi_arr,
)

RNG: Generator = np.random.default_rng()


def test_freq_lsq_freq():
    freqs = RNG.uniform(1e6, 10e9, 1000)

    assert np.allclose(lambda2_to_freq(freq_to_lambda2(freqs)), freqs)


def test_lsq_freq_lsq():
    lambda2s = RNG.uniform(1e-6, 1, 1000)

    assert np.allclose(freq_to_lambda2(lambda2_to_freq(lambda2s)), lambda2s)


def test_phi_arr():
    for max_val in [100, 1000, 10000]:
        for step_val in [1, 10, 100]:
            phi_one = make_phi_arr(max_val, step_val)
            assert len(phi_one) == max_val // step_val * 2 + 1
            assert np.isclose(np.max(phi_one), max_val)
            assert np.isclose(np.min(phi_one), -max_val)
            # Check that middle pixel is 0
            assert phi_one[len(phi_one) // 2] == 0


def test_doubel_phi_arr():
    for max_val in [100, 1000, 10000]:
        for step_val in [1, 10, 100]:
            phi_arr = make_phi_arr(max_val, step_val)
            phi_double_arr = make_double_phi_arr(phi_arr)
            assert len(phi_double_arr) == len(phi_arr) * 2 + 1
            assert np.isclose(np.max(phi_double_arr), (max_val * 2) + step_val)
            assert np.isclose(np.min(phi_double_arr), -(max_val * 2) - step_val)
            # Check that middle pixel is 0
            assert phi_double_arr[len(phi_double_arr) // 2] == 0



from __future__ import annotations

from typing import Literal, NamedTuple, Protocol

import numpy as np
import sigfig as sf
from astropy.modeling.models import Gaussian1D
from astropy.stats import akaike_info_criterion_lsq
from numpy.typing import ArrayLike, NDArray
from scipy import optimize

from rm_lite.utils.logging import logger

GAUSSIAN_SIGMA_TO_FWHM = float(2.0 * np.sqrt(2.0 * np.log(2.0)))


class StokesIModel(Protocol):
    def __call__(self, x: NDArray[np.float64], *params) -> NDArray[np.float64]: ...


class FitResult(NamedTuple):
    """Results of a Stokes I fit"""

    popt: ArrayLike
    """Best fit parameters"""
    pcov: ArrayLike
    """Covariance matrix of the fit"""
    stokes_i_model_func: StokesIModel
    """Function of the best fit model"""
    aic: float
    """Akaike Information Criterion of the fit"""


class FDFFitResult(NamedTuple):
    """Results of a Gaussian FDF fit"""

    amplitude_fit: float
    """Amplitude of the best fit model"""
    mean_fit: float
    """Mean (Faraday depth) of the best fit model"""
    stddev_fit: float
    """Standard deviation (Faraday depth) of the best fit model"""


def fwhm_to_sigma(fwhm: float) -> float:
    return float(fwhm / GAUSSIAN_SIGMA_TO_FWHM)


def sigma_to_fwhm(sigma: float) -> float:
    return float(sigma * GAUSSIAN_SIGMA_TO_FWHM)


def gaussian_integrand(
    amplitude: float,
    stddev: float | None = None,
    fwhm: float | None = None,
) -> float:
    if stddev is None and fwhm is None:
        msg = "Must provide either stddev or fwhm."
        raise ValueError(msg)
    if stddev is None and fwhm is not None:
        stddev = fwhm_to_sigma(fwhm)
    if stddev is None:
        msg = "stddev cannot be None"
        raise ValueError(msg)
    return float(amplitude * stddev * np.sqrt(2 * np.pi))


def gaussian(
    x: NDArray[np.float64],
    amplitude: float | complex,
    mean: float,
    stddev: float | None = None,
    fwhm: float | None = None,
) -> NDArray[np.float64]:
    if stddev is None and fwhm is None:
        msg = "Must provide either stddev or fwhm."
        raise ValueError(msg)
    if stddev is None and fwhm is not None:
        stddev = fwhm_to_sigma(fwhm)
    if isinstance(amplitude, complex):
        return np.array(
            Gaussian1D(amplitude=amplitude.real, mean=mean, stddev=stddev)(x)
            + 1j * Gaussian1D(amplitude=amplitude.imag, mean=mean, stddev=stddev)(x)
        )
    return np.array(Gaussian1D(amplitude=amplitude, mean=mean, stddev=stddev)(x))


def unit_gaussian(
    x: NDArray[np.float64],
    mean: float,
    stddev: float | None = None,
    fwhm: float | None = None,
) -> NDArray[np.float64]:
    if stddev is None and fwhm is None:
        msg = "Must provide either stddev or fwhm."
        raise ValueError(msg)
    if stddev is None and fwhm is not None:
        stddev = fwhm_to_sigma(fwhm)
    return Gaussian1D(amplitude=1, mean=mean, stddev=stddev)(x)


def unit_centred_gaussian(
    x: NDArray[np.float64], stddev: float | None = None, fwhm: float | None = None
) -> NDArray[np.float64]:
    if stddev is None and fwhm is None:
        msg = "Must provide either stddev or fwhm."
        raise ValueError(msg)
    if stddev is None and fwhm is not None:
        stddev = fwhm_to_sigma(fwhm)
    return Gaussian1D(amplitude=1, mean=0, stddev=stddev)(x)


def fit_rmsf(
    rmsf_to_fit_arr: NDArray[np.float64],
    phi_double_arr_radm2: NDArray[np.float64],
    fwhm_rmsf_radm2: float,
) -> float:
    rmsf_to_fit_arr = rmsf_to_fit_arr.copy()
    rmsf_to_fit_arr /= np.nanmax(rmsf_to_fit_arr)
    d_phi = phi_double_arr_radm2[1] - phi_double_arr_radm2[0]
    mask = np.zeros_like(phi_double_arr_radm2, dtype=bool)
    mask[np.argmax(rmsf_to_fit_arr)] = True
    sigma_rmsf_radm2 = fwhm_to_sigma(fwhm_rmsf_radm2)
    sigma_rmsf_arr_pix = sigma_rmsf_radm2 / d_phi
    for i in np.where(mask)[0]:
        start = int(i - sigma_rmsf_arr_pix / 2)
        end = int(i + sigma_rmsf_arr_pix / 2)
        mask[start : end + 2] = True
    popt, pcov = optimize.curve_fit(
        unit_centred_gaussian,
        phi_double_arr_radm2[mask],
        rmsf_to_fit_arr[mask],
        p0=[sigma_rmsf_radm2],
        bounds=([0], [np.inf]),
    )
    return sigma_to_fwhm(popt[0])


def fit_fdf(
    fdf_to_fit_arr: NDArray[np.float64],
    phi_arr_radm2: NDArray[np.float64],
    fwhm_fdf_radm2: float,
) -> FDFFitResult:
    d_phi = phi_arr_radm2[1] - phi_arr_radm2[0]
    mask = np.zeros_like(phi_arr_radm2, dtype=bool)
    mask[np.argmax(fdf_to_fit_arr)] = 1
    fwhm_fdf_arr_pix = fwhm_fdf_radm2 / d_phi
    fwhm_fdf_arr_pix /= 2  # fit within half the FWHM
    for i in np.where(mask)[0]:
        start = int(i - fwhm_fdf_arr_pix / 2)
        end = int(i + fwhm_fdf_arr_pix / 2)
        mask[start : end + 2] = True

    amplitude_guess = np.nanmax(fdf_to_fit_arr[mask])
    mean_guess = phi_arr_radm2[mask][np.argmax(fdf_to_fit_arr[mask])]
    stddev_guess = fwhm_fdf_radm2 / (2 * np.sqrt(2 * np.log(2)))
    popt, pcov = optimize.curve_fit(
        gaussian,
        phi_arr_radm2[mask],
        fdf_to_fit_arr[mask],
        p0=[amplitude_guess, mean_guess, stddev_guess],
    )
    logger.debug(f"Fit results: {popt}")
    amplitude_fit, mean_fit, stddev_fit = popt
    return FDFFitResult(
        amplitude_fit=amplitude_fit,
        mean_fit=mean_fit,
        stddev_fit=stddev_fit,
    )


def polynomial(order: int) -> StokesIModel:
    def poly_func(x: NDArray[np.float64], *params) -> NDArray[np.float64]:
        if len(params) != order + 1:
            msg = f"Polynomial function of order {order} requires {order + 1} parameters, {len(params)} given."
            raise ValueError(msg)
        result = 0
        for i in range(order + 1):
            result += params[i] * x**i
        return result

    return poly_func


def power_law(order: int) -> StokesIModel:
    def power_func(x: NDArray[np.float64], *params) -> NDArray[np.float64]:
        if len(params) != order + 1:
            msg = f"Power law function of order {order} requires {order + 1} parameters, {len(params)} given."
            raise ValueError(msg)
        power = 0
        for i in range(1, order + 1):
            power += params[i] * np.log10(x) ** i
        return params[0] * 10**power

    return power_func


def best_aic_func(
    aics: NDArray[np.float64], n_param: NDArray[np.int32]
) -> tuple[float, int, int]:
    """Find the best AIC for a set of AICs using Occam's razor."""
    # Find the best AIC
    best_aic_idx = int(np.nanargmin(aics))
    best_aic = float(aics[best_aic_idx])
    best_n = int(n_param[best_aic_idx])
    logger.debug(f"Lowest AIC is {best_aic}, with {best_n} params.")
    # Check if lower have diff < 2 in AIC
    aic_abs_diff = np.abs(aics - best_aic)
    bool_min_idx = np.zeros_like(aics).astype(bool)
    bool_min_idx[best_aic_idx] = True
    potential_idx = (aic_abs_diff[~bool_min_idx] < 2) & (
        n_param[~bool_min_idx] < best_n
    )
    if not any(potential_idx):
        return best_aic, best_n, best_aic_idx

    bestest_n = int(np.min(n_param[~bool_min_idx][potential_idx]))
    bestest_aic_idx = int(np.where(n_param == bestest_n)[0][0])
    bestest_aic = float(aics[bestest_aic_idx])
    logger.debug(
        f"Model within 2 of lowest AIC found. Occam says to take AIC of {bestest_aic}, with {bestest_n} params."
    )
    return bestest_aic, bestest_n, bestest_aic_idx


def static_fit(
    freq_arr_hz: NDArray[np.float64],
    ref_freq_hz: float,
    stokes_i_arr: NDArray[np.float64],
    stokes_i_error_arr: NDArray[np.float64],
    fit_order: int = 2,
    fit_type: Literal["log", "linear"] = "log",
) -> FitResult:
    msg = f"Fitting Stokes I model of type {fit_type} with order {fit_order}."
    logger.info(msg)
    if fit_type == "linear":
        fit_func = polynomial(fit_order)
    elif fit_type == "log":
        fit_func = power_law(fit_order)
    else:
        msg = f"Unknown fit type {fit_type} provided. Must be 'log' or 'linear'."  # type: ignore[unreachable]
        raise ValueError(msg)

    logger.debug(f"Fitting Stokes I model with {fit_type} model of order {fit_order}.")
    initital_guess = np.zeros(fit_order + 1)
    initital_guess[0] = np.nanmean(stokes_i_arr)
    bounds = (
        [-np.inf] * (fit_order + 1),
        [np.inf] * (fit_order + 1),
    )
    bounds[0][0] = 0.0
    if (stokes_i_error_arr == 0).all():
        stokes_i_error_arr = None
    popt, pcov = optimize.curve_fit(
        fit_func,
        freq_arr_hz / ref_freq_hz,
        stokes_i_arr,
        sigma=stokes_i_error_arr,
        absolute_sigma=True,
        p0=initital_guess,
        bounds=bounds,
    )
    stokes_i_model_arr = fit_func(freq_arr_hz / ref_freq_hz, *popt)
    ssr = np.sum((stokes_i_arr - stokes_i_model_arr) ** 2)
    with np.errstate(divide="ignore"):
        aic = akaike_info_criterion_lsq(
            ssr=ssr, n_params=fit_order + 1, n_samples=len(freq_arr_hz)
        )

    errors = np.sqrt(np.diag(pcov))
    fit_vals = [sf.round(p, e) for p, e in zip(popt, errors)]
    logger.info(f"Fit results: {fit_vals}")

    return FitResult(
        popt=popt,
        pcov=pcov,
        stokes_i_model_func=fit_func,
        aic=aic,
    )


def dynamic_fit(
    freq_arr_hz: NDArray[np.float64],
    ref_freq_hz: float,
    stokes_i_arr: NDArray[np.float64],
    stokes_i_error_arr: NDArray[np.float64],
    fit_order: int = 2,
    fit_type: Literal["log", "linear"] = "log",
) -> FitResult:
    msg = f"Iteratively fitting Stokes I model of type {fit_type} with max order {fit_order}."
    logger.info(msg)
    orders = np.arange(fit_order + 1)
    n_parameters = orders + 1
    fit_results: list[FitResult] = []

    for _i, order in enumerate(orders):
        fit_result = static_fit(
            freq_arr_hz,
            ref_freq_hz,
            stokes_i_arr,
            stokes_i_error_arr,
            order,
            fit_type,
        )
        fit_results.append(fit_result)

    logger.info(f"Fit results for orders {orders}:")
    aics = np.array([fit_result.aic for fit_result in fit_results])
    bestest_aic, bestest_n, bestest_aic_idx = best_aic_func(aics, n_parameters)
    logger.info(f"Best fit found with {bestest_n} parameters.")
    logger.debug(f"Best fit found with AIC {bestest_aic}.")
    logger.debug(f"Best fit found at index {bestest_aic_idx}.")
    logger.debug(f"Best fit found with order {orders[bestest_aic_idx]}.")

    return fit_results[bestest_aic_idx]


def fit_stokes_i_model(
    freq_arr_hz: NDArray[np.float64],
    ref_freq_hz: float,
    stokes_i_arr: NDArray[np.float64],
    stokes_i_error_arr: NDArray[np.float64],
    fit_order: int = 2,
    fit_type: Literal["log", "linear"] = "log",
) -> FitResult:
    if fit_order < 0:
        return dynamic_fit(
            freq_arr_hz,
            ref_freq_hz,
            stokes_i_arr,
            stokes_i_error_arr,
            abs(fit_order),
            fit_type,
        )

    return static_fit(
        freq_arr_hz,
        ref_freq_hz,
        stokes_i_arr,
        stokes_i_error_arr,
        fit_order,
        fit_type,
    )
